import sys

from mock import Mock, patch

from helga.bin import helga


class TestRun(object):

    def test_tcp(self):
        server = {
            'HOST': 'localhost',
            'PORT': 6667,
        }

        with patch.multiple(helga, smokesignal=Mock(), _get_comm=Mock(), reactor=Mock()):
            with patch.object(helga.settings, 'SERVER', server):
                factory = Mock()
                helga._get_comm.return_value = helga._get_comm
                helga._get_comm.Factory.return_value = factory

                helga.run()

                helga.smokesignal.emit.assert_called_with('started')
                helga.reactor.connectTCP.assert_called_with('localhost', 6667, factory)
                assert helga.reactor.run.called

    def test_ssl(self):
        server = {
            'HOST': 'localhost',
            'PORT': 6667,
            'SSL': True
        }

        with patch.multiple(helga, smokesignal=Mock(), _get_comm=Mock(), reactor=Mock(), ssl=Mock()):
            with patch.object(helga.settings, 'SERVER', server):
                ssl = Mock()
                helga.ssl.ClientContextFactory.return_value = ssl

                factory = Mock()
                helga._get_comm.return_value = helga._get_comm
                helga._get_comm.Factory.return_value = factory

                helga.run()

                helga.smokesignal.emit.assert_called_with('started')
                helga.reactor.connectSSL.assert_called_with('localhost', 6667, factory, ssl)
                assert helga.reactor.run.called


class TestMain(object):

    def test_uses_settings_env_var(monkeypatch):
        sys.argv = ['helga']

        with patch.multiple(helga, run=Mock(), settings=Mock()):
            with patch.dict('os.environ', {'HELGA_SETTINGS': 'foo'}):
                helga.main()
                helga.settings.configure.assert_called_with('foo')
                assert helga.run.called

    def test_settings_arg_overrides_env_var(self):
        sys.argv = ['helga', '--settings', 'bar']

        with patch.multiple(helga, run=Mock(), settings=Mock()):
            with patch.dict('os.environ', {'HELGA_SETTINGS': 'foo'}):
                helga.main()
                helga.settings.configure.assert_called_with('bar')
                assert helga.run.called



from mock import patch

from helga.plugins import icanhazascii


mock_animalz = {r'foo': 'bar'}


@patch('helga.plugins.icanhazascii.ANIMALZ', mock_animalz)
def test_find_animal_is_not_too_aggressive():
    assert icanhazascii.find_animal('this is a foo message') is None


@patch('helga.plugins.icanhazascii.ANIMALZ', mock_animalz)
def test_find_animal_no_match():
    assert icanhazascii.find_animal('bar') is None


@patch('helga.plugins.icanhazascii.ANIMALZ', mock_animalz)
def test_find_animal_returns_ascii_string():
    assert icanhazascii.find_animal('foo') == 'bar'


def test_plugin_flood_control():
    assert icanhazascii.icanhazascii(None, '#bots', 'me', 'foo', 'bar') is not None
    assert icanhazascii.icanhazascii(None, '#bots', 'me', 'foo', 'bar') is None



# -*- coding: utf8 -*-
from mock import patch, Mock
from pretend import stub

from helga.plugins import stfu


def test_stfu_preprocess_does_nothing():
    stfu.silenced = set()
    client = stub(nickname='helga')
    chan, nick, msg = stfu.stfu(client, '#bots', 'me', 'foo')
    assert msg == 'foo'


def test_stfu_preprocess_blanks_message_when_silenced():
    stfu.silenced = set(['#bots'])
    client = stub(nickname='helga')
    chan, nick, msg = stfu.stfu(client, '#bots', 'me', 'foo')
    assert msg == ''


def test_stfu_snark_on_private_message():
    stfu.silenced = set()
    resp = stfu.stfu(stub(nickname='helga'), 'me', 'me', 'helga stfu', 'stfu', [])
    assert resp in map(lambda x: x.format(nick='me'), stfu.snarks)


def test_stfu_silences_channel():
    stfu.silenced = set()
    resp = stfu.stfu(stub(nickname='helga'), '#bots', 'me', 'helga stfu', 'stfu', [])
    assert resp in map(lambda x: x.format(nick='me'), stfu.silence_acks)
    assert '#bots' in stfu.silenced


@patch('helga.plugins.stfu.reactor')
def test_stfu_for_some_time(reactor):
    client = stub(nickname='helga')
    stfu.stfu(client, '#bots', 'me', 'helga stfu for 30', 'stfu', ['for', '30'])
    reactor.callLater.assert_called_with(30*60, stfu.auto_unsilence, client, '#bots', 30*60)


def test_stfu_unsilences_channel():
    stfu.silenced = set(['#bots'])
    resp = stfu.stfu(stub(nickname='helga'), '#bots', 'me', 'helga speak', 'speak', [])
    assert resp in map(lambda x: x.format(nick='me'), stfu.unsilence_acks)
    assert '#bots' not in stfu.silenced


def test_stfu_speak_only_speaks_once():
    stfu.silenced = set(['#bots'])
    resp = stfu.stfu(stub(nickname='helga'), '#bots', 'me', 'helga speak', 'speak', [])
    assert resp in map(lambda x: x.format(nick='me'), stfu.unsilence_acks)

    resp = stfu.stfu(stub(nickname='helga'), '#bots', 'me', 'helga speak', 'speak', [])
    assert resp is None


def test_auto_unsilence():
    stfu.silenced = set(['#bots'])
    client = Mock()
    stfu.auto_unsilence(client, '#bots', 300)
    assert '#bots' not in stfu.silenced
    client.msg.assert_called_with('#bots', 'Speaking again after waiting 5 minutes')


def test_stfu_with_unicode():
    client = stub(nickname=u'☃')
    chan, nick, msg = stfu.stfu(client, '#bots', 'me', u'helga speak ☃')
    assert chan == '#bots'
    assert nick == 'me'
    assert msg == u'helga speak ☃'


@patch('helga.plugins.stfu.reactor')
def test_stfu_handles_invalid_args(reactor):
    client = stub(nickname='helga')
    test_args = [
        ['for'],
        ['for', 'blah'],
        ['for', None],
    ]

    for args in test_args:
        stfu.stfu(client, '#bots', 'me', 'helga stfu for 30', 'stfu', args)
        assert not reactor.callLater.called



# coding: utf-8
import pytest

from mock import Mock

from helga.webhooks import logger


class TestIndexView(object):

    def setup(self):
        self.view = logger.Index()

    def test_title(self):
        assert self.view.title() == 'Channel Logs'

    def test_channels(self, monkeypatch):
        monkeypatch.setattr(logger, 'os', Mock())
        logger.os.listdir.return_value = ['#foo', '#bar', '#baz']

        assert list(self.view.channels()) == ['bar', 'baz', 'foo']

    def test_channels_empty_for_no_logs(self, monkeypatch):
        monkeypatch.setattr(logger, 'os', Mock())
        logger.os.path.isdir.return_value = False
        logger.os.listdir.side_effect = OSError

        try:
            retval = list(self.view.channels())
        except:
            pytest.fail('Should not have raised an Exception')
        else:
            assert retval == []

    def test_channels_hides_blacklist(self, monkeypatch):
        monkeypatch.setattr(logger, 'os', Mock())

        # Should handle with or without leading '#'
        monkeypatch.setattr(logger, 'settings', Mock(
            CHANNEL_LOGGING_HIDE_CHANNELS=['#foo', 'bar']
        ))

        logger.os.listdir.return_value = ['#foo', '#bar', '#baz']

        assert list(self.view.channels()) == ['baz']


class TestChannelIndexView(object):

    def setup(self):
        self.view = logger.ChannelIndex('foo')

    def test_title(self):
        assert self.view.title() == '#foo Channel Logs'

    def test_dates(self, monkeypatch):
        monkeypatch.setattr(logger, 'os', Mock())
        logger.os.path.isdir.return_value = True
        logger.os.listdir.return_value = [
            '2010-12-01.txt',
            '2011-12-01.txt',
            '2012-12-01.txt',
            '2012-10-31.txt',
        ]

        assert list(self.view.dates()) == [
            '2012-12-01',
            '2012-10-31',
            '2011-12-01',
            '2010-12-01',
        ]

    def test_dates_404(self, monkeypatch):
        monkeypatch.setattr(logger, 'os', Mock())
        logger.os.path.isdir.return_value = False

        with pytest.raises(logger.HttpError):
            list(self.view.dates())


class TestChannelLogView(object):

    def setup(self):
        self.view = logger.ChannelLog('foo', '2014-12-01')

    def test_title(self):
        assert self.view.title() == '#foo Channel Logs for 2014-12-01'

    def test_messages_404(self, monkeypatch):
        monkeypatch.setattr(logger, 'os', Mock())
        logger.os.path.isfile.return_value = False

        with pytest.raises(logger.HttpError):
            list(self.view.messages())

    def test_messages(self, tmpdir):
        logger.settings.CHANNEL_LOGGING_DIR = str(tmpdir)

        # Create tmp file
        file = tmpdir.mkdir('#foo').join('2014-12-01.txt')
        file.write('\n'.join([
            '00:00:00 - foo - this is what i said',
            '12:01:35 - bar - another thing i said',
            '16:17:18 - baz - this - has - delimiters',
            u'21:22:23 - qux - ☃'.encode('utf-8')
        ]), mode='wb')

        assert list(self.view.messages()) == [
            {
                'time': '00:00:00',
                'nick': 'foo',
                'message': 'this is what i said',
            },
            {
                'time': '12:01:35',
                'nick': 'bar',
                'message': 'another thing i said',
            },
            {
                'time': '16:17:18',
                'nick': 'baz',
                'message': 'this - has - delimiters',
            },
            {
                'time': '21:22:23',
                'nick': 'qux',
                'message': u'☃',
            },
        ]

    def test_messages_with_multiline_content(self, tmpdir):
        logger.settings.CHANNEL_LOGGING_DIR = str(tmpdir)

        # Create tmp file
        file = tmpdir.mkdir('#foo').join('2014-12-01.txt')
        file.write('\n'.join([
            '00:00:00 - foo - this is what i said',
            '...and here',
            '...and here again',
            u'...☃'.encode('utf-8')
        ]), mode='wb')

        assert list(self.view.messages()) == [
            {
                'time': '00:00:00',
                'nick': 'foo',
                'message': u'this is what i said\n...and here\n...and here again\n...☃',
            }
        ]

    def test_messages_with_unhandled_content(self, tmpdir):
        logger.settings.CHANNEL_LOGGING_DIR = str(tmpdir)

        # Create tmp file
        file = tmpdir.mkdir('#foo').join('2014-12-01.txt')
        file.write("it's lonely here")

        assert list(self.view.messages()) == [
            {
                'time': '',
                'nick': '',
                'message': "it's lonely here",
            }
        ]

    def test_download(self, tmpdir):
        request = Mock()
        contents = ('00:00:00 - foo - this is what i said\n'
                    '12:01:35 - bar - another thing i said\n'
                    '16:17:18 - baz - this - has - delimiters\n'
                    u'21:22:23 - qux - ☃'.encode('utf-8'))

        logger.settings.CHANNEL_LOGGING_DIR = str(tmpdir)

        # Create tmp file
        file = tmpdir.mkdir('#foo').join('2014-12-01.txt')
        file.write(contents, mode='wb')

        assert self.view.download(request) == contents
        request.setHeader.assert_any_call('Content-Type', 'text/plain')
        request.setHeader.assert_any_call('Content-Disposition', 'attachment; filename=2014-12-01.txt')


class TestWebhook(object):

    def setup(self):
        self.request = Mock()
        logger.settings.CHANNEL_LOGGING = True

    def test_raises_501(self):
        logger.settings.CHANNEL_LOGGING = False
        with pytest.raises(logger.HttpError):
            logger.logger(None, None)

    def _mock_log_dir(self, tmpdir):
        # Create tmp file
        logger.settings.CHANNEL_LOGGING_DIR = str(tmpdir)
        file = tmpdir.mkdir('#foo').join('2014-12-01.txt')
        file.write('00:00:00 - foo - this is what i said')

    def test_renders_index(self, tmpdir):
        self._mock_log_dir(tmpdir)
        response = logger.logger(self.request, None)
        self.request.setHeader.assert_called_with('Content-Type', 'text/html')

        # Output asserts
        assert '<title>Channel Logs</title>' in response
        assert '<a href="/logger/foo">#foo</a>' in response

    def test_renders_channel_index(self, tmpdir):
        self._mock_log_dir(tmpdir)
        response = logger.logger(self.request, None, 'foo')
        self.request.setHeader.assert_called_with('Content-Type', 'text/html')

        # Output asserts
        assert '<title>#foo Channel Logs</title>' in response
        assert '<a href="/logger/foo/2014-12-01">2014-12-01</a>' in response

    def test_renders_channel_log(self, tmpdir):
        self._mock_log_dir(tmpdir)
        response = logger.logger(self.request, None, 'foo', '2014-12-01')
        self.request.setHeader.assert_called_with('Content-Type', 'text/html')

        # Output asserts
        assert '<title>#foo Channel Logs for 2014-12-01</title>' in response
        assert '<td><a href="#00:00:00" name="00:00:00">00:00:00</a></td>' in response
        assert '<td>foo</td>' in response
        assert '<td><pre>this is what i said</pre></td>' in response

    def test_renders_channel_log_as_text(self, tmpdir):
        self._mock_log_dir(tmpdir)
        response = logger.logger(self.request, None, 'foo', '2014-12-01', as_text=True)
        self.request.setHeader.assert_any_call('Content-Type', 'text/plain')
        self.request.setHeader.assert_any_call('Content-Disposition', 'attachment; filename=2014-12-01.txt')

        # Output asserts
        assert response == '00:00:00 - foo - this is what i said'



"""
`pymongo`_ connection objects and utilities

.. attribute:: client

    A :class:`pymongo.mongo_client.MongoClient` instance, the connection client to MongoDB

.. attribute:: db

    A :class:`pymongo.database.Database` instance, the default MongoDB database to use


.. _`pymongo`: http://api.mongodb.org/python/current/
"""
import warnings


from pymongo import MongoClient
from pymongo.errors import ConnectionFailure

from helga import settings


def connect():
    """
    Connect to a MongoDB instance, if helga is configured to do so (see setting
    :data:`~helga.settings.DATABASE`). This will return the MongoDB client as well
    as the default database as configured.

    :returns: A two-tuple of (:class:`pymongo.MongoClient`, :class:`pymongo.database.Database`)
    """
    db_settings = getattr(settings, 'DATABASE', {})

    try:
        client = MongoClient(db_settings['HOST'], db_settings['PORT'])
    except ConnectionFailure:
        warnings.warn('MongoDB is not available. Some features may not work')
        return None, None
    else:
        db = client[db_settings['DB']]

        if 'USERNAME' in db_settings and 'PASSWORD' in db_settings:
            db.authenticate(db_settings['USERNAME'], db_settings['PASSWORD'])

        return client, db


client, db = connect()



from giphypop import GiphyApiException
from mock import Mock, patch
from pretend import stub

from helga.plugins import giphy


mock_api = Mock()
mock_api.return_value = mock_api
mock_img = stub(media_url='GIF')


@patch('helga.plugins.giphy.Giphy', mock_api)
def test_gifme_returns_random_gif():
    mock_api.return_value = mock_api
    mock_api.random_gif.return_value = mock_img

    resp = giphy.giphy(None, '#bots', 'me', 'helga gifme foo', 'gifme', ['foo'])
    assert mock_api.random_gif.called
    assert resp == mock_img.media_url


@patch('helga.plugins.giphy.Giphy', mock_api)
def test_gifme_returns_translated():
    mock_api.return_value = mock_api
    mock_api.random_gif.side_effect = GiphyApiException
    mock_api.translate.return_value = mock_img

    resp = giphy.giphy(None, '#bots', 'me', 'helga gifme foo', 'gifme', ['foo'])
    assert mock_api.translate.called
    assert resp == mock_img.media_url


@patch('helga.plugins.giphy.Giphy', mock_api)
def test_gifme_returns_searched():
    mock_api.random_gif.side_effect = GiphyApiException
    mock_api.translate.side_effect = GiphyApiException
    mock_api.search_list.return_value = [mock_img]

    resp = giphy.giphy(None, '#bots', 'me', 'helga gifme foo', 'gifme', ['foo'])
    assert mock_api.search_list.called
    assert resp == mock_img.media_url


@patch('helga.plugins.giphy.Giphy', mock_api)
def test_gifme_returns_snark():
    mock_api.random_gif.side_effect = GiphyApiException
    mock_api.translate.side_effect = GiphyApiException
    mock_api.search_list.side_effect = GiphyApiException

    formatted = map(lambda x: x.format(nick='me'), giphy.responses)

    assert giphy.giphy(None, '#bots', 'me', 'helga gifme foo', 'gifme', ['foo']) in formatted



from collections import defaultdict

from helga.plugins import command, registry


def format_help_string(name, *helps):
    return u'[{0}] {1}'.format(name, '. '.join(helps))


@command('help', aliases=['halp'],
         help="Show the help string for any commands. Usage: helga help [<plugin>]")
def help(client, channel, nick, message, cmd, args):
    helps = defaultdict(list)
    default_help = u'No help string for this plugin'

    for plugin_name in registry.enabled_plugins[channel]:
        try:
            plugin = registry.plugins[plugin_name]
        except KeyError:
            continue

        # A simple object
        if hasattr(plugin, 'help'):
            helps[plugin_name].append(plugin.help or default_help)

        # A decorated function
        elif hasattr(plugin, '_plugins'):
            fn_helps = filter(bool, map(lambda x: getattr(x, 'help', None), plugin._plugins))
            helps[plugin_name].extend(fn_helps or [default_help])

    try:
        plugin = args[0]
    except IndexError:
        pass
    else:
        if plugin not in registry.enabled_plugins[channel]:
            return u"Sorry {0}, I don't know about that plugin".format(nick)
        elif plugin not in helps.keys():
            return u"Sorry {0}, there's no help string for plugin '{1}'".format(nick, plugin)

        # Single plugin, it's probably ok in the public channel
        return format_help_string(plugin, *helps[plugin])

    if channel != nick:
        client.me(channel, 'whispers to {0}'.format(nick))

    retval = []
    # Send the message to the user
    for key, value in helps.iteritems():
        retval.append(format_help_string(key, *value))

    retval.insert(0, u"{0}, here are the plugins I know about".format(nick))
    client.msg(nick, u'\n'.join(retval))



"""
Utilities for interacting with twitter

.. note::

    This module may be deprecated in the future
"""
import tweepy

from helga import log, settings


logger = log.getLogger(__name__)


def is_properly_configured():
    """
    Ensures that all necessary settings for communicating with twitter are configured.
    This includes:

    * ``TWITTER_CONSUMER_KEY``
    * ``TWITTER_CONSUMER_SECRET``
    * ``TWITTER_OAUTH_TOKEN``
    * ``TWITTER_OAUTH_TOKEN_SECRET``

    :returns: boolean, True if all settings are set
    """
    return all([
        getattr(settings, 'TWITTER_CONSUMER_KEY', None),
        getattr(settings, 'TWITTER_CONSUMER_SECRET', None),
        getattr(settings, 'TWITTER_OAUTH_TOKEN', None),
        getattr(settings, 'TWITTER_OAUTH_TOKEN_SECRET', None),
    ])


def message_max(message, max):
    """
    Returns a given message with a maximum length. Primarily used for additional logging

    :param str message: a message intended to send to twitter
    :param int max: the maximum length of the message
    :returns: the message truncated to a maximum length
    """
    if len(message) > max:
        logger.warning('Message exceeds %s characters. Truncating', max)
        message = message[:max]
    return message


def message_140(message):
    """
    A wrapper for message_max using 140 as a hard limit

    :param str message: a message intended to send to twitter
    :returns: the message truncated to a maximum length of 140 characters
    """
    return message_max(message, 140)


def get_api():
    """
    Constructs a ``tweepy.API`` object using configured twitter settings in ``helga.settings``.

    :returns: an instance of ``tweepy.API``
    """
    auth = tweepy.OAuthHandler(settings.TWITTER_CONSUMER_KEY, settings.TWITTER_CONSUMER_SECRET)
    auth.set_access_token(settings.TWITTER_OAUTH_TOKEN, settings.TWITTER_OAUTH_TOKEN_SECRET)
    return tweepy.API(auth)


def tweet(message):
    """
    Sends a tweet using ``tweepy``. Ensures that helga is properly configured for accessing
    twitter and sends a message with a maximum character length of 140 characters.

    :param str message: the message to tweet
    :returns: a URL string of the posted tweet
    """
    if not is_properly_configured():
        logger.error('Twitter API requires consumer key, consumer secret, oauth token, oauth secret')
        return

    message = message_140(message)

    try:
        logger.info('Tweeting: %s', message)
        status = get_api().update_status(message)
    except:  # pragma: no cover
        logger.exception('Could not post status')
    else:
        tweet_url = 'http://twitter.com/{0}/status/{1}'.format(settings.TWITTER_USERNAME, status.id)
        logger.info('Tweeted: %s', tweet_url)
        return tweet_url



import random

from helga.plugins import match


descriptors = ('Abstract', 'Base', 'Lazy', 'Mutable', 'Composite', 'Recursive',
               'Asynchronous', 'Shared', 'Runnable', 'Secure', 'Generic', 'Internal',
               'External', 'Prioritized',
               )

frameworks = ('Swing', 'Hibernate', 'JDBC', 'Spring', 'Struts', 'Android', 'AWT',
              '', '', '', ''  # Additional weight for random choice
              )

things = ('Printer', 'Internet', 'Synergy', 'IRC', 'Window', 'Network', 'TCP', 'Protocol',
          'Data', 'Web', 'Script', 'Command', 'Television', 'UML', 'Null', 'Scheduler',
          'Bandwidth', 'Bot', 'Computation', 'Event', 'Color', 'Font', 'Logger', 'Hash',
          'SOAP', 'XML', 'JSON', 'HTTP', 'REST', 'Layout', 'Cache', 'List', 'Enum', 'Device',
          'Statement', 'Memory', 'Timestamp', 'Graph', 'Result', 'Grid', 'Category', 'Story',
          'Article', 'Model', 'Tokenizer',
          )

actions = ('Extraction', 'Mutation', 'Automation', 'Interpolation', 'Instantiation',
           'Translation', 'Iteration', '', '', '', '',  # Additional weight for random choice
           )

oop = ('Thread', 'Bean', 'Command', 'Prototype', 'Factory', 'Builder', 'Singleton', 'Client',
       'Exception', 'Error', 'Pool', 'Decorator', 'Controller', 'Proxy', 'Module', 'Adapter',
       'Bridge', 'Wrapper', 'Iterator', 'Observer', 'Interpreter', 'Applet', 'Action', 'Map',
       'Queue', 'Configuration', 'Encoder', 'Tokenizer',
       )


@match(r'(lol)?java(?!script)', priority=0)
def make_bullshit_java_thing(client, channel, nick, message, matches):
    """
    Java class naming is hilariously enterprise...
    """
    return (random.choice(descriptors) +
            random.choice(frameworks) +
            random.choice(things + oop) +
            random.choice(actions) +
            random.choice(oop))



import random

from giphypop import Giphy, GiphyApiException, GIPHY_PUBLIC_KEY

from helga import log, settings
from helga.plugins import command


logger = log.getLogger(__name__)

# Snarky responses on failure
responses = [
    "Well this is embarassing...",
    "Yeah I've got nothing {nick}",
    "I couldn't find anything for you {nick}",
    "PC LOAD LETTER",
]


@command('gif', aliases=['gifme'], help='Search giphy for an animated gif. Usage: helga gif <search term>')
def giphy(client, channel, nick, message, cmd, args):
    key = getattr(settings, 'GIPHY_API_KEY', GIPHY_PUBLIC_KEY)
    api = Giphy(api_key=key, strict=True)

    search = ' '.join(args)

    try:
        return api.random_gif(search).media_url
    except GiphyApiException:
        try:
            return api.translate(search).media_url
        except GiphyApiException:
            try:
                return api.search_list(search, limit=1)[0].media_url
            except (GiphyApiException, IndexError):
                pass

    return random.choice(responses).format(nick=nick)



from helga import settings
from helga.plugins import command


@command('showme', aliases=['whois', 'whothehellis'],
         help="Show a URL for the user's intranet page. Usage: helga (showme|whois|whothehellis) <nick>")
def wiki_whois(client, channel, nick, message, cmd, args):  # pragma: no cover
    """
    Show the intranet page for a user. Settings must have a WIKI_URL value with formattable
    substring named {user}
    """
    return settings.WIKI_URL.format(user=args[0])



# -*- coding: utf8 -*-
from mock import Mock, patch

from helga import settings
from helga.util import twitter


def test_is_properly_configured():
    settings.TWITTER_CONSUMER_KEY = 'foo'
    settings.TWITTER_CONSUMER_SECRET = 'foo'
    settings.TWITTER_OAUTH_TOKEN = 'foo'
    settings.TWITTER_OAUTH_TOKEN_SECRET = 'foo'

    assert twitter.is_properly_configured()

    settings.TWITTER_CONSUMER_KEY = ''
    assert not twitter.is_properly_configured()

    delattr(settings, 'TWITTER_CONSUMER_KEY')
    assert not twitter.is_properly_configured()


def test_message_max():
    msg1 = 'a' * 255
    msg2 = 'a'
    assert len(twitter.message_max(msg1, 140)) == 140
    assert len(twitter.message_max(msg2, 140)) == 1


def test_message_max_handles_unicode():
    snowman1 = u'☃' * 255
    snowman2 = u'☃'
    assert len(twitter.message_max(snowman1, 140)) == 140
    assert len(twitter.message_max(snowman2, 140)) == 1


@patch('helga.util.twitter.get_api')
def test_tweet(api):
    settings.TWITTER_CONSUMER_KEY = 'foo'
    settings.TWITTER_CONSUMER_SECRET = 'foo'
    settings.TWITTER_OAUTH_TOKEN = 'foo'
    settings.TWITTER_OAUTH_TOKEN_SECRET = 'foo'
    settings.TWITTER_USERNAME = 'helgabot'

    api.return_value = api
    api.update_status.return_value = Mock(id=123456789)

    assert twitter.tweet('foo') == 'http://twitter.com/helgabot/status/123456789'


@patch('helga.util.twitter.get_api')
def test_tweet_handles_unicode(api):
    settings.TWITTER_CONSUMER_KEY = 'foo'
    settings.TWITTER_CONSUMER_SECRET = 'foo'
    settings.TWITTER_OAUTH_TOKEN = 'foo'
    settings.TWITTER_OAUTH_TOKEN_SECRET = 'foo'
    settings.TWITTER_USERNAME = 'helgabot'

    api.return_value = api
    api.update_status.return_value = Mock(id=123456789)

    assert twitter.tweet(u'☃') == 'http://twitter.com/helgabot/status/123456789'


@patch('helga.util.twitter.message_140')
@patch('helga.util.twitter.is_properly_configured')
def test_tweet_with_improperly_configured_settings(configured, message_140):
    configured.return_value = False
    twitter.tweet('foobar')
    assert not message_140.called


@patch('helga.util.twitter.tweepy')
@patch('helga.util.twitter.settings')
def test_get_api(settings, tweepy):
    settings.TWITTER_CONSUMER_KEY = 'foo'
    settings.TWITTER_CONSUMER_SECRET = 'foo'
    settings.TWITTER_OAUTH_TOKEN = 'foo'
    settings.TWITTER_OAUTH_TOKEN_SECRET = 'foo'
    settings.TWITTER_USERNAME = 'helgabot'

    tweepy.OAuthHandler.return_value = tweepy
    tweepy.API.return_value = tweepy

    assert twitter.get_api() == tweepy

    tweepy.OAuthHandler.assert_called_with(settings.TWITTER_CONSUMER_KEY,
                                           settings.TWITTER_CONSUMER_SECRET)
    tweepy.set_access_token.assert_called_with(settings.TWITTER_OAUTH_TOKEN,
                                               settings.TWITTER_OAUTH_TOKEN_SECRET)
    tweepy.API.assert_called_with(tweepy)



# -*- coding: utf8 -*-
import random
import re

from helga.plugins import match


def imgur(image):
    """
    Returns an imgur link with a given hash
    """
    return 'http://i.imgur.com/{0}.gif'.format(image)


RESPONSES = {
    # Direct text responses
    r'(gross|disgusting|eww)': (imgur('XEEI0Rn'),),  # Dumb and Dumber Gag

    r'(\sGFY\s|GTFO|Fuck (You|Off))': (imgur('VPqgYjF'),    # Ryan Stiles pulling middle finger from pocket
                                       imgur('rWhZY3k'),),  # half baked

    r'womp womp': ("http://www.sadtrombone.com/?play=true",
                   "http://www.youtube.com/watch?v=_-GaXa8tSBE"),

    r'^:w?q$': ("this ain't your vi",),

    r'^(pwd$|(sudo|ls|cd|rm)(\s\w+|$))': "this ain't your shell",

    r'php': ("php is just terrible",
             "php's motto: MERGE ALL THE PULL REQUESTS"),

    r'^select( .* )from(.*)': "'; DROP TABLES;",

    r'mongo(db)?\s': 'http://youtu.be/b2F-DItXtZs',  # MongoDB is webscale

    r'gem install': "ruby. not even once.",

    r'^(logger|logs)\?$': "http://logger.ddtc.cmgdigital.com/",

    r'^docs\?$': 'http://docs.cmgdigital.com/',

    r'\\m/': 'rock on',

    r'((beetle|betel)(geuse|juice)\s?){3}': "i'm the ghost with the most",

    # lol, gifs
    r'(bravo|well done)': (imgur('wSvsV'),     # Citizen Kane slow clap
                           imgur('HUKCsCv'),   # Colbert & Stewart bravo
                           imgur('FwqHZ6Z')),  # Gamer conceding defeat

    r'is \w+ down\?': imgur('yX5o8rZ'),  # THE F5 HAMMER

    r"(i don't care|do i look like i care|zero fucks)": (
        imgur('oKydfNm'),  # Bird bouncing on hawk's head
        imgur('KowlC'),    # Gangam style 'do i look like i care'
        imgur('xYOqXJv'),  # Dog hitting cat with tail
        imgur('1b2YNU3'),  # But wait! bubble
    ),

    r'^nope$': (imgur('iSm1aZu'),   # Arrested development NOPE
                imgur('2xwe756'),   # Lonley island like a boss NOPE
                imgur('zCtbl'),     # Tracy Morgan NOPE
                imgur('foEHo'),     # Spongebob buried in sand
                imgur('xKYs9'),     # Puppy does not like lime
                imgur('ST9lw3U'),   # Seinfeld I'm Out
                imgur('c4gTe5p'),   # Cat thriller walk I'm out
                'http://i.minus.com/iUgVCKwjISSke.gif',  # The Nope Bader
                ),

    r'tl;?dr': (imgur('dnMjc'),   # Lightsaber did not read
                imgur('V2H9y')),  # Craig Robinson did not read

    r'panic': (imgur('tpGQV'),    # Aladding start panicking
               imgur('WS4S2'),    # Colbert screaming in terror
               imgur('rhNOy3I'),  # Panic cat bug eyes
               imgur('SNvM6CZ'),  # Girl leans on escalator handrail
               imgur('H7PXV'),    # Ain't nobody got time for that
               imgur('fH9e2')),   # Out of control truck on collision course

    r'shock(ed|ing)?': (imgur('zVyOBlR'),   # Cartoon is shocked
                        imgur('Q4bI5'),     # Shocked cat is shocked
                        imgur('wdA2Z'),     # Monsters Inc watching Boo in compactor
                        imgur('nj3yp'),     # Spock is shocked
                        imgur('AGnOQ'),     # PeeWee is shocked
                        imgur('wkY1FUI'),   # Shocked looks around
                        imgur('AXuUYIj')),  # Simpsons jaw drop

    r'(bloody mary|vodka)': imgur('W9SS4iJ'),  # Archer: Bloody Mary, blessed are you among cocktails

    r'popcorn': (imgur('00IJgSZ'),  # Thriller popcorn
                 imgur('5px9l')),   # Colbert popcorn

    r'deal with it': (imgur('12WoH'),    # Slip n slide DWI
                      imgur('6E6n1'),    # WTF Oprah
                      imgur('hYdy4'),    # Baseball catch deal with it
                      imgur('pqmfX'),    # WTF pouring water from nose
                      imgur('9WbAL'),    # A three toed sloth in a chair
                      imgur('KdldmZk'),  # Polar bear jumping out of water
                      imgur('49UtI5N'),  # The Fresh Prince of DEAL WITH IT
                      imgur('1pkNeOy'),  # Skyler
                      imgur('KzEXQDq'),  # Tom & Jerry
                      imgur('1kxk9z6'),  # deal with it dance
                      u'(⌐■_■)',

                      # Multiline
                      (u'( •_•)',
                       u'( •_•)>⌐■-■',
                       u'(⌐■_■)',
                       'deal with it'),

                      (u'. B     :-|',
                       u'.   B   :-|',
                       u'.     B :-|',
                       u'.       B-| deal with it')),

    r'(mind blown|blew my mind)': (imgur('U6kCXUp'),   # Head asploding
                                   imgur('1HMveGj')),  # Tim and Eric mind blown

    r'(sweet jesus|mother of god)': (imgur('5vXdAOV'),   # Captain Kirk
                                     imgur('g155Wra'),   # Star Trek freaking out
                                     imgur('dyeHb'),     # BJ Novak looks confused
                                     imgur('VkHiG6D'),   # Face twitching
                                     imgur('aiH4Mts'),   # Christopher Lloyd realizes something
                                     imgur('nOJme'),     # Cookie monster sweet jesus
                                     imgur('KtdHWhs'),   # Fight club realization
                                     imgur('z5hhSsU'),   # Cat with toy: OMG it was you!
                                     imgur('zuc9tAm')),  # Dinosaurs show - drops beer

    r'nailed it': (imgur('KsQzQTF'),   # Cat not trying to catch rat
                   imgur('5nrEk'),     # Olympic diving fail
                   imgur('n9zw0'),     # squirrel spinning on bird feeder
                   imgur('puZy04m'),   # Kid jumping into pool fail
                   imgur('MBdxv'),     # Girl trying to jump bike ramp fail
                   imgur('6XRqt'),     # FIXME
                   imgur('dFuBE'),     # Cat jumps into a box
                   imgur('vUACp'),     # Backflip off bleachers
                   imgur('59h9A8e')),  # Backflip off tree

    r'unacceptable': imgur('BwdP2xl'),  # 3D rendering goes wrong

    r'^(iknorite|right)\?$': imgur('RvquHs0'),  # Breaking Bad: You're god damn right

    r'fuck yea': (imgur('GZ5CD5r'),   # Data shooting dice
                  imgur('nEmrMkq')),  # Top Gun ... DANGER ZONE

    r'\w+ broke prod': (imgur('SuCGnum'),  # Anchorman: You ate the whole wheel of cheese?
                        imgur('sbQUDbF'),),  # fail boat

    r'^indeed$': imgur('bQcbpki'),  # Leonardo DiCaprio in Django Unchained

    r'f(f{6}|7)u(u{11}|12)': 'http://i.minus.com/ibnfJRQi1h4z30.gif',  # Workaholics: FUUUUUUUUUUUUUU

    r'wtf': imgur('bpW6Xkd'),  # WTF supercut


    # Various modern unicode emoticons
    r'(why|y) (u|you) no':                      u'ლ(ಠ益ಠლ)',
    r'i (don\'?t know|dunno),? lol':            u'¯\(°_o)/¯',
    r'look.?of.?disapproval(\.jpg|\.gif)?':     u'ಠ_ಠ',
    r'i (am disappoint|disapprove)':            u'ಠ_ಠ',
    r'^not sure if \w+':                        u'≖_≖',

    r'(tableflip|flip (a|the|some) tables?)':   (u'(╯°□°）╯︵ ┻━┻',
                                                 u'(ノಠ益ಠ)ノ彡┻━┻'),

    r'(gonna|going to) (make \w+ )?cry':        u'(ಥ﹏ಥ)',
    r'(bro ?fist|fist ?bump)':                  u'( _)=mm=(^_^ )',

    r'hi(gh)?[ -]?five':                        ('\o',
                                                 u'( ‘-’)人(ﾟ_ﾟ )'),

    r'(^|[^\\])o/$':                            '\o',
    r'^\\o$':                                   'o/'
}


def find_response(message):
    for pat, resp in RESPONSES.iteritems():
        if re.findall(pat, message, re.I):
            found = resp
            break
    else:
        return None

    if isinstance(resp, (tuple, list)):
        return random.choice(found)
    return found


@match(find_response, priority=0)
def oneliner(client, channel, nick, message, match):
    """
    Maybe some of these will become their own thing, but for
    now, they live here.

    DEAL WITH IT
    """
    return match  # pragma: no cover



from datetime import datetime as dt

__title__ = 'helga'
__version__ = '1.6.8'
__author__ = 'Shaun Duncan'
__license__ = 'MIT/GPLv3'
__copyright__ = 'Copyright {0} Shaun Duncan'.format(dt.now().strftime('%Y'))
__description__ = 'A full-featured IRC bot for Python/Twisted with plugin support'



"""
Webhook HTTP server plugin and core webhook API

Webhooks provide a way to expose HTTP endpoints that can interact with helga. A command
plugin manages an HTTP server that is run on a port specified by setting
:data:`helga.settings.WEBHOOKS_PORT` (default 8080). An additional, optional setting that can be
used for routes requiring HTTP basic auth is :data:`helga.settings.WEBOOKS_CREDENTIALS`,
which should be a list of tuples, where each tuple is a pair of (username, password).

Routes are URL path endpoints. On the surface they are just python callables decorated using
:func:`@route <route>`. The route decorated must be given a path regex, and optional list of
HTTP methods to accept. Webhook plugins must be registered in the same way normal plugins are
registered, using setuptools entry_points. However, they must belong to the entry_point group
``helga_webhooks``. For example::

    setup(entry_points={
        'helga_webhooks': [
            'api = myapi.decorated_route'
        ]
    })

For more information, see :ref:`webhooks`
"""
import functools
import pkg_resources
import re

from twisted.internet import reactor
from twisted.web import server, resource
from twisted.web.error import Error

import smokesignal

from helga import log, settings
from helga.plugins import Command, registry
from helga.util.encodings import from_unicode


logger = log.getLogger(__name__)


# Subclassed only for better naming
class HttpError(Error):
    pass


class WebhookPlugin(Command):
    """
    A command plugin that manages running an HTTP server for webhook routes and services. Usage::

        helga webhooks (start|stop|routes)

    Both ``start`` and ``stop`` are privileged actions and can start and stop the HTTP listener for
    webhooks respectively. To use them, a user must be configured as an operator. The ``routes``
    subcommand will list all of the URL routes known to the webhook listener.
    """
    command = 'webhooks'
    help = ('HTTP service for interacting with helga. Command options usage: '
            'helga webhooks (start|stop|routes). Note: start/stop'
            'can be run only by helga operators')

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)

        # Per issue 137, these were previously set on signon, but there is a bit of a
        # chicken and egg situation where routes in the same module as commands and matches
        # would try to register themselves before the connection was made, so the command
        # or match would fail to load.
        self.root = WebhookRoot()
        self.site = server.Site(self.root)
        self.port = getattr(settings, 'WEBHOOKS_PORT', 8080)

        @smokesignal.on('signon')
        def setup(client):  # pragma: no cover
            self._start(client)
            self._init_routes()

    def _init_routes(self):
        enabled = settings.ENABLED_WEBHOOKS

        for entry_point in pkg_resources.iter_entry_points(group='helga_webhooks'):
            if enabled is not None and entry_point.name not in enabled:
                logger.info('Skipping disabled webhook %s', entry_point.name)
                continue

            try:
                logger.info('Loading webhook %s', entry_point.name)
                entry_point.load()
            except:  # pragma: no cover
                logger.exception('Error loading webhook %s', entry_point)

    def _start(self, client=None):
        logger.info('Starting webhooks service on port %s', self.port)
        self.root.irc_client = client
        self.tcp = reactor.listenTCP(self.port, self.site)

    def _stop(self):
        logger.info('Stopping webhooks service on port %s', self.port)
        self.tcp.stopListening()
        self.tcp.loseConnection()
        self.tcp = None

    def add_route(self, fn, path, methods):
        """
        Adds a route handler function to the root web resource at a given path
        and for the given methods

        :param fn: the route handler function
        :param str path: the URL path of the route
        :param list methods: list of HTTP methods that the route should respond to
        """
        self.root.add_route(fn, path, methods)  # pragma: no cover

    def list_routes(self, client, nick):
        """
        Messages a user with all webhook routes and their supported HTTP methods

        :param client: an instance of :class:`helga.comm.Client`
        :param str nick: the nick of the IRC user to message
        """
        client.msg(nick, u'{0}, here are the routes I know about'.format(nick))
        for pattern, route in self.root.routes.iteritems():
            http_methods = route[0]  # Route is a tuple (http_methods, function)
            client.msg(nick, u'[{0}] {1}'.format(','.join(http_methods), pattern))

    def control(self, action):
        """
        Control the running HTTP server. Intended for helga operators.

        :param str action: the action to perform, either 'start' or 'stop'
        """
        running = self.tcp is not None

        if action == 'stop':
            if running:
                self._stop()
                return u"Webhooks service stopped"
            return u"Webhooks service not running"

        if action == 'start':
            if not running:
                self._start()
                return u"Webhooks service started"
            return u"Webhooks service already running"

    def run(self, client, channel, nick, msg, cmd, args):
        try:
            subcmd = args[0]
        except IndexError:
            subcmd = 'routes'

        if subcmd == 'routes':
            client.me(channel, u'whispers to {0}'.format(nick))
            self.list_routes(client, nick)
        elif subcmd in ('start', 'stop'):
            if nick not in client.operators:
                return u"Sorry {0}, Only an operator can do that".format(nick)
            return self.control(subcmd)


class WebhookRoot(resource.Resource):
    """
    The root HTTP resource the webhook HTTP server uses to respond to requests. This
    manages all registered webhook route handlers, manages running them, and manages
    returning any responses generated.
    """
    isLeaf = True

    def __init__(self, *args, **kwargs):
        #: An instance of :class:`helga.comm.Client`
        self.irc_client = None

        #: A dictionary of regular expression URL paths as keys, and two-tuple values
        #: of allowed methods, and the route handler function
        self.routes = {}

    def add_route(self, fn, path, methods):
        """
        Adds a route handler function to the root web resource at a given path
        and for the given methods

        :param fn: the route handler function
        :param str path: the URL path of the route
        :param list methods: list of HTTP methods that the route should respond to
        """
        self.routes[path] = (methods, fn)

    def render(self, request):
        """
        Renders a response for an incoming request. Handles finding and dispatching the route
        matching the incoming request path. Any response string generated will be explicitly
        encoded as a UTF-8 byte string.

        If no route patch matches the incoming request, a 404 is returned.

        If a route is found, but the request uses a method that the route handler does not
        support, a 405 is returned.

        :param request: The incoming HTTP request, ``twisted.web.http.Request``
        :returns: a string with the HTTP response content
        """
        request.setHeader('Server', 'helga')
        for pat, route in self.routes.iteritems():
            match = re.match(pat, request.path)
            if match:
                break
        else:
            request.setResponseCode(404)
            return '404 Not Found'

        # Ensure that this route handles the request method
        methods, fn = route
        if request.method.upper() not in methods:
            request.setResponseCode(405)
            return '405 Method Not Allowed'

        # Handle raised HttpErrors
        try:
            # Explicitly return a byte string. Twisted expects this
            return from_unicode(fn(request, self.irc_client, **match.groupdict()))
        except HttpError as e:
            request.setResponseCode(e.status)
            return e.message or e.response


def authenticated(fn):
    """
    Decorator for declaring a webhook route as requiring HTTP basic authentication.
    Incoming requests validate a supplied basic auth username and password against the list
    configured in the setting :data:`~helga.settings.WEBHOOKS_CREDENTIALS`. If no valid
    credentials are supplied, an HTTP 401 response is returned.

    :param fn: the route handler to decorate
    """
    @functools.wraps(fn)
    def ensure_authenticated(request, *args, **kwargs):
        for user, password in getattr(settings, 'WEBHOOKS_CREDENTIALS', []):
            if user == request.getUser() and password == request.getPassword():
                return fn(request, *args, **kwargs)

        # No valid basic auth provided
        request.setResponseCode(401)
        return '401 Unauthorized'
    return ensure_authenticated


def route(path, methods=None):
    """
    Decorator to register a webhook route. This requires a path regular expression, and
    optionally a list of HTTP methods to accept, which defaults to accepting ``GET`` requests
    only. Incoming HTTP requests that use a non-allowed method will receive a 405 HTTP response.

    :param str path: a regular expression string for the URL path of the route
    :param methods: a list of accepted HTTP methods for this route, defaulting to ``['GET']``

    Decorated routes must follow this pattern:

    .. function:: func(request, client)
        :noindex:

        :param request: The incoming HTTP request, ``twisted.web.http.Request``
        :param client: The IRC client connection. An instance of :class:`helga.comm.Client`
        :returns: a string HTTP response
    """
    plugin = registry.get_plugin('webhooks')
    if methods is None:
        methods = ['GET']

    def wrapper(fn):
        if plugin is not None:
            plugin.add_route(fn, path, methods)
        return fn

    return wrapper



"""
Helga's core plugin library containing base implementations for creating plugins
as well as utilities for managing plugins at runtime
"""
from __future__ import absolute_import
import functools
import pkg_resources
import random
import re
import shlex
import sys
import warnings

from collections import defaultdict
from itertools import ifilter, imap
from operator import methodcaller

import smokesignal

from helga import log, settings
from helga.util.encodings import from_unicode, to_unicode


logger = log.getLogger(__name__)


#: A collection of pre-canned acknowledgement type responses
ACKS = [
    'roger',
    '10-4',
    'no problem',
    'will do',
    'you got it',
    'anything you say',
    'sure thing',
    'ok',
    'right-o',
    'consider it done',
]


#: The value for low priority plugins.
#: Configurable via setting :data:`~helga.settings.PLUGIN_PRIORITY_LOW`
PRIORITY_LOW = settings.PLUGIN_PRIORITY_LOW

#: The value for normal priority plugins.
#: Configurable via setting :data:`~helga.settings.PLUGIN_PRIORITY_NORMAL`
PRIORITY_NORMAL = settings.PLUGIN_PRIORITY_NORMAL

#: The value for high priority plugins.
#: Configurable via setting :data:`~helga.settings.PLUGIN_PRIORITY_HIGH`
PRIORITY_HIGH = settings.PLUGIN_PRIORITY_HIGH


if not settings.COMMAND_ARGS_SHLEX:
    warnings.warn(u'Command arg parsing will default to shlex in a future version', FutureWarning)


def random_ack():
    """
    Returns a random choice from :data:`ACKS`
    """
    return random.choice(ACKS)


class ResponseNotReady(Exception):
    """
    Exception raised by plugins that perform some async operation using
    twisted deferreds. If the bot is configured to only allow the first plugin
    response (by default), then any plugin raising this will prevent further
    plugin execution

    (see :ref:`plugins.async`)
    """


class Registry(object):
    """
    Simple plugin registry that handles dispatching messages to registered plugins.
    Plugins can be disabled or enabled per channel. By default, plugins are loaded, but not
    enabled on a channel unless it exists in :data:`~helga.settings.ENABLED_PLUGINS`. This is done so that
    potentially annoying plugins can be enabled on-demand

    .. attribute:: plugins
        :annotation: = {}

        A dictionary mapping plugin names to decorated functions or :class:`Plugin` subclasses

    .. attribute:: enabled_plugins
        :annotation: = {}

        A dictionary of enabled plugin names per channel, keyed by channel name
    """
    __instance = None

    def __new__(cls, *args, **kwargs):
        """
        Only here so we only maintain one registry for the life of the application. There
        is some state-specific things that shouldn't be lost in the event the IRC client
        loses the connection to the server
        """
        if cls.__instance is None:
            cls.__instance = super(Registry, cls).__new__(cls, *args, **kwargs)
        return cls.__instance

    def __init__(self):
        if not hasattr(self, 'plugins'):
            self.plugins = {}

        if not hasattr(self, 'enabled_plugins'):
            # Enabled plugins is a dict: channel -> set()
            self.enabled_plugins = defaultdict(lambda: set(getattr(settings, 'ENABLED_PLUGINS', [])))

        smokesignal.on('started', self.load)

    def register(self, name, fn_or_cls):
        """
        Register a decorated plugin function or :class:`Plugin` subclass with a given name

        :param str name: the name of the plugin
        :param fn_or_cls: a decorated plugin function or :class:`Plugin` subclass
        :raises TypeError: if the ``fn_or_cls`` argument is not a decorated plugin function or
                           :class:`Plugin` subclass
        """
        # Make sure we're working with an instance
        try:
            if issubclass(fn_or_cls, Plugin):
                fn_or_cls = fn_or_cls()
        except TypeError:
            pass

        if not (isinstance(fn_or_cls, Plugin) or hasattr(fn_or_cls, '_plugins')):
            raise TypeError(u"Plugin {0} must be a subclass of Plugin, or a decorated function".format(name))

        logger.info('Registered plugin %s', name)
        self.plugins[name] = fn_or_cls

    @property
    def all_plugins(self):
        """
        A set of all registered plugin names
        """
        return set(self.plugins.keys())

    def get_plugin(self, name):
        """
        Get a plugin by name

        :param str name: the name of the plugin
        :returns: a plugin implementation (decorated function or :class:`Plugin` sublclass)
        """
        return self.plugins.get(name, None)

    def disable(self, channel, *plugins):
        """
        Disable a plugin or plugins on a desired channel

        :param str channel: the desired IRC channel
        :param \*plugins: a list of plugin names to disable
        """
        self.enabled_plugins[channel] = self.enabled_plugins[channel].difference(set(plugins))

    def enable(self, channel, *plugins):
        """
        Enable a plugin or plugins on a desired channel

        :param str channel: the desired IRC channel
        :param \*plugins: a list of plugin names to enable
        """
        self.enabled_plugins[channel] = self.enabled_plugins[channel].union(set(plugins))

    def load(self):
        """
        Load all plugins registered via setuptools entry point named ``helga_plugins`` and
        initialize them. For example::

            entry_points = {
                'helga_plugins': [
                    'plugin_name = mylib.mymodule:MyPluginClass',
                ],
            }

        """
        for entry_point in pkg_resources.iter_entry_points(group='helga_plugins'):
            try:
                logger.debug('Loading entry_point %s', entry_point.name)
                self.register(entry_point.name, entry_point.load())
            except:
                logger.exception('Error initializing plugin %s', entry_point)
        smokesignal.emit('plugins_loaded')

    def reload(self, name):
        """
        Reloads a plugin with a given name. This is equivalent to finding the registered
        entry point module and using the python builtin ``reload()``.

        :param str name: the desired plugin to reload
        :returns: True if reloaded, False if an exception occurred
        """
        if name not in self.plugins:
            # FIXME: This should raise
            return u"Unknown plugin '{0}'. Is it installed?".format(name)

        for entry_point in pkg_resources.iter_entry_points(group='helga_plugins'):
            if entry_point.name != name:
                continue

            # FIXME: exceptions should bubble up
            try:
                reload(sys.modules[entry_point.module_name])
                self.register(entry_point.name, entry_point.load())
                return True
            except:
                logger.exception('Failed to reload plugin %s', entry_point)
                return False

    def prioritized(self, channel, high_to_low=True):
        """
        Obtain a list of enabled plugins for a given channel ordered according to their priority
        (see :ref:`plugins.priorities`). The default action is to return a list ordered from most
        important to least important.

        :param str channel: the IRC channel for the enabled plugin list
        :param bool high_to_low: priority ordering, True for most important to least important.
        """
        plugins = []
        for name in self.enabled_plugins[channel]:
            if name not in self.plugins:
                logger.debug('Plugin %s may not be installed or have incorrect entry_point information', name)
                continue

            # Decorated functions will have this
            if hasattr(self.plugins[name], '_plugins'):
                plugins.extend(self.plugins[name]._plugins)
            else:
                plugins.append(self.plugins[name])

        return sorted(plugins, key=lambda p: getattr(p, 'priority', PRIORITY_NORMAL), reverse=high_to_low)

    def preprocess(self, client, channel, nick, message):
        """
        Invoke the ``preprocess`` method for each plugin on a given channel according to plugin priority.
        Any exceptions from plugins will be suppressed and logged.

        :param client: an instance of :class:`helga.comm.Client`
        :param str channel: the channel from which the message came
        :param str nick: the nick of the user sending the message
        :param str message: the original message received
        :returns: a three-tuple (channel, nick, message) containing modifications all preprocessor
                  plugins have made
        """
        for plugin in self.prioritized(channel):
            try:
                channel, nick, message = plugin.preprocess(client, channel, nick, message)
            except:
                logger.exception('Calling preprocess on plugin %s failed', plugin)
                continue

        return channel, nick, message

    def process(self, client, channel, nick, message):
        """
        Invoke the ``process`` method for each plugin on a given channel according to plugin priority.
        Any exceptions from plugins will be suppressed and logged. All return values from plugin
        ``process`` methods are collected unless the setting
        :data:`~helga.settings.PLUGIN_FIRST_RESPONDER_ONLY` is set to True or a plugin raises
        :exc:`ResponseNotReady`, in which case the first plugin to return a response or raise
        :exc:`ResponseNotReady` will prevent others from processing. All response strings are explicitly
        converted to unicode.

        :param client: an instance of :class:`helga.comm.Client`
        :param str channel: the channel from which the message came
        :param str nick: the nick of the user sending the message
        :param str message: the original message received
        :returns: a list of non-empty unicode response strings
        """
        responses = []
        first_responder = getattr(settings, 'PLUGIN_FIRST_RESPONDER_ONLY', False)

        for plugin in self.prioritized(channel):
            try:
                resp = plugin.process(client, channel, nick, message)
            except ResponseNotReady:
                if first_responder:
                    break
                continue
            except:
                logger.exception('Calling process on plugin %s failed', plugin)
                continue

            if not resp:
                continue

            # Chained decorator style plugins return a list of strings
            if isinstance(resp, (tuple, list)):
                # Be sure to filter Nones, then strip
                responses.extend(imap(lambda s: (s or '').strip(), resp))
            else:
                responses.append(resp.strip())

            if responses and first_responder:
                break

        # FIXME: Explicit conversion to unicode might not make sense. Perpahs
        # a warning should be sent to the user? Or do we even care?
        return map(to_unicode, ifilter(bool, responses))


registry = Registry()


class Plugin(object):
    """
    The base class for helga plugins. There are three main methods of this base class that are
    important for creating class-based plugins.

    ``preprocess``

    Run by the plugin registry as a preprocessing mechanism. Allows plugins to modify
    the channel, nick, and/or message that other plugins will receive.

    ``process``

    Run by the plugin registry to allow a plugin to process an IRC message. This is
    the primary entry point for plugins according to the plugin manager, so it should either return
    a response or not.

    ``run``

    Run internally by the plugin, generally from within the ``process`` method. This should
    do the actual work to generate a response. In other words, ``process`` should handle
    checking if the plugin should handle a message and then return whatever ``run`` returns.
    """
    #: The registered priority of the plugin
    priority = PRIORITY_NORMAL

    def __init__(self, priority=PRIORITY_NORMAL):
        self.priority = priority

    def run(self, client, channel, nick, message, *args, **kwargs):
        """
        Executes this plugin with a given message to generate a response. This should run without
        regard to whether it should or should not for a given message. Note, that this is where the
        actual work for the plugin should occur. Subclasses should implement this method.

        A return value of None, an empty string, or empty list implies that no response should be
        sent over IRC. A non-empty string, list of strings, or raised :exc:`ResponseNotReady` implies
        a response to be sent.

        :param client: The IRC client connection. An instance of :class:`helga.comm.Client`
        :param channel: The channel from which the message was received. This could be a public
                        channel like '#foo', or in the event of a private message, could be the
                        nick of the user sending the message
        :param nick: The nick of the user sending the message
        :param message: The full message string received from IRC
        :returns: None if no response is to be sent back over IRC, a non-empty string or list
                  of strings if a response is to be returned
        """
        return None  # pragma: no cover

    def preprocess(self, client, channel, nick, message):
        """
        A preprocessing filter for plugins. This allows a plugin to modify a received
        message prior to that message being handled by this plugin's or other plugin's
        ``process`` method.

        :param client: The IRC client connection. An instance of :class:`helga.comm.Client`
        :param channel: The channel from which the message was received. This could be a public
                        channel like '#foo', or in the event of a private message, could be the
                        nick of the user sending the message
        :param nick: The nick of the user sending the message
        :param message: The full message string received from IRC
        :returns: a three-tuple (channel, nick, message) containing any modifications
        """
        return channel, nick, message  # pragma: no cover

    def process(self, client, channel, nick, message):
        """
        This method of a plugin is called by helga's plugin registry to process an incoming
        IRC message. This should determine whether or not the plugin ``run`` method should be
        called. If so, it should return whatever return value ``run`` generates. If not, ``None``
        should be returned.

        :param client: The IRC client connection. An instance of :class:`helga.comm.Client`
        :param channel: The channel from which the message was received. This could be a public
                        channel like '#foo', or in the event of a private message, could be the
                        nick of the user sending the message
        :param nick: The nick of the user sending the message
        :param message: The full message string received from IRC
        :returns: None if the plugin should not run, otherwise the return value of the ``run`` method
        """
        return self.run(client, channel, nick, message)

    def decorate(self, fn, preprocessor=False):
        """
        A helper for decorating a function to handle this plugin. This essentially just monkey
        patches the ``run`` or ``preprocess`` method whith the given function. Decorated functions
        should accept whatever arguments the subclass implementation sends to its ``run`` method.
        Also, instances/subclasses of :class:`Plugin` are kept in a list attribute of the decorated
        function. This allows chainable decorators that function as intended. Example usage::

            def my_plugin(*args, **kwargs):
                pass

            p = Plugin()
            p.decorate(my_plugin)
            assert p in my_plugin._plugins

        :param fn: function to decorate
        :param bool preprocessor: True if the function should be decorated as a preprocessor
        """
        if preprocessor:
            self.preprocess = fn
        else:
            self.run = fn

        try:
            fn._plugins.append(self)
        except AttributeError:
            fn._plugins = [self]

        return fn

    def __call__(self, client, channel, nick, message):
        """
        Proxy for ``process``
        """
        return self.process(client, channel, nick, message)


class Command(Plugin):
    """
    A subclass of :class:`Plugin` for command type plugins (see :ref:`plugins.types`). Command
    plugins have a default priority of :data:`PRIORITY_NORMAL`
    """

    #: The command string, i.e. 'search' for a command 'helga search foo'
    command = ''

    #: A list of command aliases. If a command 'search' has an alias list ['s'], then
    #: 'helga search foo' and 'helga s foo' will both run the command
    aliases = []

    #: An optional help string for the command. This is used by the builtin
    #: :ref:`builtin.plugins.help` plugin
    help = ''

    #: A boolean indicating whether or not to use shlex arg string parsing rather than naive
    #: whitespace splitting
    shlex = False

    def __init__(self, command='', aliases=None, help='', priority=PRIORITY_NORMAL, shlex=False):
        super(Command, self).__init__(priority)
        self.command = command or self.command
        self.aliases = aliases or self.aliases
        self.help = help or self.help
        self.shlex = shlex

    def parse(self, botnick, message):
        """
        Parse the incoming message using the current nick of the bot, the defined command string of
        this object, plus any aliases. Will return the actual command parsed (which could be an alias),
        plus either whitespaced delimited list of strings that follow the parsed command, or shlex
        argument list if ``shlex`` is True.

        Generally, this does not need to be implemented by subclasses

        :param botnick: the current bot nickname
        :param message: the incoming IRC message
        :returns: two-tuple consisting of the string of parsed command, and an argument list of
                  strings either whitespace delimited or shlex split.
        """
        choices = [self.command] + list(self.aliases)

        # Sort choices from longest to shortest. This will ease a quirk where
        # short alias versions will trump the more verbose ones
        choices = sorted(choices, key=len, reverse=True)

        nick_prefix = ''

        # Handle multiple ways to parse this command
        if getattr(settings, 'COMMAND_PREFIX_BOTNICK', True):
            nick_prefix = '{0}\W*\s'.format(botnick)

        prefixes = filter(bool, [nick_prefix, getattr(settings, 'COMMAND_PREFIX_CHAR', '!')])
        prefix = '({0})'.format('|'.join(prefixes))

        pat = ur'^{0}({1})($|\s(.*)$)'.format(prefix, '|'.join(choices))

        try:
            _, cmd, _, argstr = re.findall(pat, message, re.IGNORECASE)[0]
        except (IndexError, ValueError):
            # FIXME: Log here?
            return u'', []

        return cmd, filter(bool, self._parse_argstr(argstr))

    def _parse_argstr(self, argstr):
        """
        Parse an argument string for this command. If COMMAND_ARGS_SHLEX is set to False,
        then naive whitespace splitting is performed on the argument string. If not, a more robust
        ``shlex.split()`` is performed. For example, given the message::

            helga foo bar "baz qux"

        the former would produce arguments::

            ['bar', '"baz', 'qux"']

        while the latter would produce::

            ['bar', 'baz qux']

        """
        if self.shlex or settings.COMMAND_ARGS_SHLEX:
            argv = shlex.split(from_unicode(argstr.strip()))
        else:
            argv = argstr.strip().split(' ')

        return map(to_unicode, argv)

    def run(self, client, channel, nick, message, command, args):
        """
        Executes this plugin with a given message to generate a response. This should run without
        regard to whether it should or should not for a given message. Note, that this is where the
        actual work for the plugin should occur. Subclasses should implement this method.

        A return value of None, an empty string, or empty list implies that no response should be
        sent over IRC. A non-empty string, list of strings, or raised :exc:`ResponseNotReady` implies
        a response to be sent.

        :param client: The IRC client connection. An instance of :class:`helga.comm.Client`
        :param channel: The channel from which the message was received. This could be a public
                        channel like '#foo', or in the event of a private message, could be the
                        nick of the user sending the message
        :param nick: The nick of the user sending the message
        :param message: The full message string received from IRC
        :param cmd: The parsed command string which could be the registered command or one of the
                    command aliases
        :param args: The parsed command arguments as a list, i.e. any content following the command.
                     For example: ``helga foo bar baz`` would be ``['bar', 'baz']``
        :returns: String or list of strings to return over IRC. None or empty string or list for no response
        """
        return None  # pragma: no cover

    def process(self, client, channel, nick, message):
        """
        Parses the incoming message and determins if this command should run (i.e. if the primary
        command or one of the aliases match).

        :param client: The IRC client connection. An instance of :class:`helga.comm.Client`
        :param channel: The channel from which the message was received. This could be a public
                        channel like '#foo', or in the event of a private message, could be the
                        nick of the user sending the message
        :param nick: The nick of the user sending the message
        :param message: The full message string received from IRC
        :returns: None if the plugin should not run, otherwise the return value of the ``run`` method
        """
        command, args = self.parse(client.nickname, message)
        all_commands = [self.command] + list(self.aliases)

        if settings.COMMAND_IGNORECASE:
            command = command.lower()
            all_commands = map(methodcaller('lower'), all_commands)

        if command not in all_commands:
            return None

        return self.run(client, channel, nick, message, command, args)


class Match(Plugin):
    """
    A subclass of :class:`Plugin` for match type plugins (see :ref:`plugins.types`). Matches
    have a default priority of :data:`PRIORITY_LOW`
    """
    #: A regular expression string used to match against an IRC message. Optionally, this attribute can
    #: be a callable that accepts an IRC message string as its only argument and returns a value that
    #: can be evaluated for truthiness.
    pattern = ''

    def __init__(self, pattern='', priority=PRIORITY_LOW):
        super(Match, self).__init__(priority)
        self.pattern = pattern or self.pattern

    def run(self, client, channel, nick, message, matches):
        """
        Executes this plugin with a given message to generate a response. This should run without
        regard to whether it should or should not for a given message. Note, that this is where the
        actual work for the plugin should occur. Subclasses should implement this method.

        A return value of None, an empty string, or empty list implies that no response should be
        sent over IRC. A non-empty string, list of strings, or raised :exc:`ResponseNotReady` implies
        a response to be sent.

        :param client: The IRC client connection. An instance of :class:`helga.comm.Client`
        :param channel: The channel from which the message was received. This could be a public
                        channel like '#foo', or in the event of a private message, could be the
                        nick of the user sending the message
        :param nick: The nick of the user sending the message
        :param message: The full message string received from IRC
        :param matches: The result of ``re.findall`` if decorated with a regular expression, otherwise
                        the return value of the callable passed to :func:`helga.plugins.match`
        :returns: String or list of strings to return over IRC. None or empty string or list for no response
        """
        return None  # pragma: no cover

    def match(self, message):
        """
        Matches a message against the pattern defined for this class. If the ``pattern`` attribute
        is a callable, it is called with the message as its only argument and that value is returned.
        Otherwise, the pattern attribute is used as a regular expression string argument to
        ``re.findall`` and that value is returned.

        :param message: the message received over IRC
        :returns: the result of ``re.findall`` if pattern is a string, otherwise the return value of
                  calling the ``pattern`` attribute with the message as a parameter
        """
        if callable(self.pattern):
            fn = self.pattern
        else:
            fn = functools.partial(re.findall, self.pattern)

        try:
            return fn(message)
        except TypeError:
            return None

    def process(self, client, channel, nick, message):
        """
        Processes a message sent by a user on a given channel. This will return None if the message does
        not match the plugin's pattern, or the return value of ``run`` if it does match. For this plugin
        to match an incoming message, the return value of ``self.match()`` must return value that can
        be evaluated for truth. Generally, subclasses should not have to worry about this method, and
        instead, should focus on the implementation of ``run``.

        :param client: The IRC client connection. An instance of :class:`helga.comm.Client`
        :param channel: The channel from which the message was received. This could be a public
                        channel like '#foo', or in the event of a private message, could be the
                        nick of the user sending the message
        :param nick: The nick of the user sending the message
        :param message: The full message string received from IRC
        :returns: None if the plugin should not run, otherwise the return value of the ``run`` method
        """
        matches = self.match(message)
        if not bool(matches):
            return None

        return self.run(client, channel, nick, message, matches)


def command(command, aliases=None, help='', priority=PRIORITY_NORMAL, shlex=False):
    """
    A decorator for creating command plugins

    :param str command: The command string, i.e. 'search' for a command 'helga search foo'
    :param list aliases: A list of command aliases. If a command 'search' has an alias list
                         ['s'], then 'helga search foo' and 'helga s foo' will both run the command.
    :param str help: An optional help string for the command. This is used by the builtin help plugin.
    :param int priority: The priority of the plugin. Default is :data:`PRIORITY_NORMAL`.
    :param bool shlex: A boolean indicating whether to use shlex arg string parsing rather than naive
                       whitespace splitting.

    Decorated functions should follow this pattern:

    .. function:: func(client, channel, nick, message, cmd, args)
        :noindex:

        :param client: The IRC client connection. An instance of :class:`helga.comm.Client`
        :param channel: The channel from which the message was received. This could be a public
                        channel like '#foo', or in the event of a private message, could be the
                        nick of the user sending the message
        :param nick: The nick of the user sending the message
        :param message: The full message string received from IRC
        :param cmd: The parsed command string which could be the registered command or one of the
                    command aliases
        :param args: The parsed command arguments as a list, i.e. any content following the command.
                     For example: ``helga foo bar baz`` would be ``['bar', 'baz']``
        :returns: String or list of strings to return over IRC. None or empty string or list
                  for no response
    """
    return Command(command, aliases=aliases, help=help, priority=priority, shlex=shlex).decorate


def match(pattern, priority=PRIORITY_LOW):
    """
    A decorator for creating match plugins

    :param pattern: A regular expression string used to match against an IRC message. Optionally,
                    this argument can be a callable that accepts an IRC message string as its only
                    argument and returns a value that can be evaluated for truthiness.
    :param int priority: The priority of the plugin. Default is :data:`PRIORITY_LOW`

    Decorated match functions should follow this pattern:

    .. function:: func(client, channel, nick, message, matches)
        :noindex:

        :param client: The IRC client connection. An instance of :class:`helga.comm.Client`
        :param channel: The channel from which the message was received. This could be a public
                        channel like '#foo', or in the event of a private message, could be the
                        nick of the user sending the message
        :param nick: The nick of the user sending the message
        :param message: The full message string received from IRC
        :param matches: The result of ``re.findall`` if decorated with a regular expression, otherwise
                        the return value of the callable passed
        :returns: String or list of strings to return over IRC. None or empty string or list for no response
    """
    return Match(pattern, priority=priority).decorate


def preprocessor(priority=PRIORITY_NORMAL):
    """
    A decorator for creating preprocessor plugins

    :param int priority: The priority of the plugin. Default is :data:`PRIORITY_NORMAL`

    Decorated preprocessor functions should follow this pattern:

    .. function:: func(client, channel, nick, message, matches)
        :noindex:

        :param client: The IRC client connection. An instance of :class:`helga.comm.Client`
        :param channel: The channel from which the message was received. This could be a public
                        channel like '#foo', or in the event of a private message, could be the
                        nick of the user sending the message
        :param nick: The nick of the user sending the message
        :param message: The full message string received from IRC
        :returns: a three-tuple (channel, nick, message) containing any modifications
    """
    # This happens if not using a priority argument, but just decorating
    if callable(priority):
        return Plugin(priority=PRIORITY_NORMAL).decorate(priority, preprocessor=True)
    else:
        return functools.partial(Plugin(priority=priority).decorate, preprocessor=True)



import os
import sys

import pytest

from helga import settings


def test_configure_execfile(tmpdir):
    file = tmpdir.join('foo.py')
    file.write('MY_CUSTOM_SETTING = "foo"')

    assert not hasattr(settings, 'MY_CUSTOM_SETTING')
    settings.configure(str(file))
    assert settings.MY_CUSTOM_SETTING == 'foo'
    delattr(settings, 'MY_CUSTOM_SETTING')


def test_configure_import(tmpdir):
    file = tmpdir.join('foo.py')
    file.write('MY_CUSTOM_SETTING = "foo"')

    sys.path.insert(0, str(tmpdir))
    filename = os.path.basename(str(file)).replace('.py', '')

    assert not hasattr(settings, 'MY_CUSTOM_SETTING')
    settings.configure(filename)
    assert settings.MY_CUSTOM_SETTING == 'foo'
    delattr(settings, 'MY_CUSTOM_SETTING')


def test_configure_import_raises():
    with pytest.raises(ImportError):
        settings.configure('foo.bar')



import os
import re

from collections import deque
from itertools import imap
from operator import methodcaller

import pystache

from helga import settings
from helga.plugins.webhooks import HttpError, route
from helga.util.encodings import to_unicode


class Index(object):
    """
    Rendered object for the logger index page meant to show the full list
    of logged channels.
    """

    def title(self):
        return u'Channel Logs'

    def channels(self):
        log_dir = settings.CHANNEL_LOGGING_DIR
        lstrip = methodcaller('lstrip', '#')
        hidden = set(imap(lstrip, settings.CHANNEL_LOGGING_HIDE_CHANNELS))

        if not os.path.isdir(log_dir):
            raise StopIteration

        for chan in imap(lstrip, sorted(os.listdir(log_dir))):
            if chan in hidden:
                continue
            yield chan


class ChannelIndex(object):
    """
    Rendered object for the logger channel index page meant to show the full list
    of log files (UTC dates) for a given IRC channel.
    """

    def __init__(self, channel):
        self.channel = channel

    def title(self):
        return u'#{0} Channel Logs'.format(self.channel)

    def dates(self):
        channel = '#{0}'.format(self.channel)
        basedir = os.path.join(settings.CHANNEL_LOGGING_DIR, channel)

        if not os.path.isdir(basedir):
            raise HttpError(404)

        for date in sorted(os.listdir(basedir), reverse=True):
            yield date.replace('.txt', '')


class ChannelLog(object):
    """
    Rendered object for displaying the full contents of a channel log for a
    given channel and date.
    """

    def __init__(self, channel, date):
        self.channel_name = channel
        self.date = date
        self.logfile = '{0}.txt'.format(self.date)
        self.channel = '#{0}'.format(self.channel_name)

    @property
    def logfile_path(self):
        return os.path.join(settings.CHANNEL_LOGGING_DIR, self.channel, self.logfile)

    def title(self):
        """
        The page title
        """
        return u'{0} Channel Logs for {1}'.format(self.channel, self.date)

    def messages(self):
        """
        Generator for logged channel messages as a dictionary
        of the message time, message nick, and message contents
        """
        if not os.path.isfile(self.logfile_path):
            raise HttpError(404)

        line_pat = re.compile(r'^(\d{2}:?){3} - \w+ - .*$')
        message = u''
        log = deque()

        # XXX: This is kind of terrible. Some things will log only a single time
        # if the message sent over IRC has newlines. So we have to read in reverse
        # and construct the response list
        with open(self.logfile_path, 'r') as fp:
            for line in imap(to_unicode, reversed(fp.readlines())):
                if not line_pat.match(line):
                    message = u''.join((line, message))
                    continue

                parts = line.strip().split(u' - ')
                time = parts.pop(0)
                nick = parts.pop(0)
                message = u'\n'.join((u' - '.join(parts), message))
                log.appendleft({
                    'time': time,
                    'nick': nick,
                    'message': message.rstrip(u'\n'),
                })
                message = ''

        if message:
            log.appendleft({
                'time': '',
                'nick': '',
                'message': message,
            })

        return log

    def download(self, request):
        """
        Offers this logfile as a download
        """
        request.setHeader('Content-Type', 'text/plain')
        request.setHeader('Content-Disposition',
                          'attachment; filename={0}'.format(self.logfile))
        with open(self.logfile_path, 'r') as fp:
            return '\n'.join(line.strip() for line in fp.readlines())


@route(r'/logger/?$')
@route(r'/logger/(?P<channel>[\w\-_]+)/?$')
@route(r'/logger/(?P<channel>[\w\-_]+)/(?P<date>[\w\-]+)(?P<as_text>\.txt)?/?$')
def logger(request, irc_client, channel=None, date=None, as_text=None):
    if not settings.CHANNEL_LOGGING:
        raise HttpError(501, 'Channel logging is not enabled')

    request.setHeader('Content-Type', 'text/html')
    renderer = pystache.renderer.Renderer(
        search_dirs=os.path.dirname(os.path.abspath(__file__))
    )

    if channel is None:
        page = Index()
    elif date is None:
        page = ChannelIndex(channel)
    else:
        page = ChannelLog(channel, date)
        if as_text is not None:
            return page.download(request)

    return renderer.render(page)



# -*- coding: utf8 -*-
from unittest import TestCase

from mock import Mock, patch, call

from helga.plugins import webhooks
from twisted.web import server


@patch('helga.plugins.webhooks.registry')
def test_route(reg):
    reg.get_plugin.return_value = reg
    fake_fn = lambda: 'foo'
    webhooks.route('/foo', methods=['GET', 'POST'])(fake_fn)

    reg.add_route.assert_called_with(fake_fn, '/foo', ['GET', 'POST'])


@patch('helga.plugins.webhooks.registry')
def test_route_with_no_methods(reg):
    reg.get_plugin.return_value = reg
    fake_fn = lambda: 'foo'
    webhooks.route('/foo')(fake_fn)

    reg.add_route.assert_called_with(fake_fn, '/foo', ['GET'])


@patch('helga.plugins.webhooks.settings')
def test_authenticated_passes(settings):
    @webhooks.authenticated
    def fake_fn(*args, **kwargs):
        return 'OK'

    settings.WEBHOOKS_CREDENTIALS = [('foo', 'bar')]

    request = Mock()
    request.getUser.return_value = 'foo'
    request.getPassword.return_value = 'bar'

    assert fake_fn(request) == 'OK'


@patch('helga.plugins.webhooks.settings')
def test_authenticated_fails_when_called(settings):
    @webhooks.authenticated
    def fake_fn(*args, **kwargs):
        return 'OK'

    settings.WEBHOOKS_CREDENTIALS = [('person', 'password')]

    request = Mock()
    request.getUser.return_value = 'foo'
    request.getPassword.return_value = 'bar'

    assert fake_fn(request) == '401 Unauthorized'
    request.setResponseCode.assert_called_with(401)


class WebhookPluginTestCase(TestCase):

    def setUp(self):
        self.plugin = webhooks.WebhookPlugin()

    def test_initializes_root_and_site(self):
        plugin = webhooks.WebhookPlugin()
        assert isinstance(plugin.root, webhooks.WebhookRoot)
        assert isinstance(plugin.site, webhooks.server.Site)

    @patch('helga.plugins.webhooks.settings')
    def test_custom_port(self, settings):
        settings.WEBHOOKS_PORT = 1337
        plugin = webhooks.WebhookPlugin()
        assert self.plugin.port == 8080
        assert plugin.port == 1337

    @patch('helga.plugins.webhooks.settings')
    @patch('helga.plugins.webhooks.pkg_resources')
    def test_init_routes(self, pkg_resources, settings):
        settings.ENABLED_WEBHOOKS = None
        entry_points = [Mock()]
        pkg_resources.iter_entry_points.return_value = entry_points
        self.plugin._init_routes()
        pkg_resources.iter_entry_points.assert_called_with(group='helga_webhooks')
        assert entry_points[0].load.called

    @patch('helga.plugins.webhooks.settings')
    @patch('helga.plugins.webhooks.pkg_resources')
    def test_init_routes_ignores_not_enabled(self, pkg_resources, settings):
        settings.ENABLED_WEBHOOKS = ['foo']
        entry_points = [Mock(), Mock()]
        entry_points[0].name = 'foo'
        entry_points[1].name = 'bar'
        pkg_resources.iter_entry_points.return_value = entry_points
        self.plugin._init_routes()
        assert entry_points[0].load.called
        assert not entry_points[1].load.called

    @patch('helga.plugins.webhooks.reactor')
    def test_start(self, reactor):
        client = Mock()
        self.plugin._start(client)
        assert isinstance(self.plugin.root, webhooks.WebhookRoot)
        assert isinstance(self.plugin.site, server.Site)
        reactor.listenTCP.assert_called_with(8080, self.plugin.site)

    @patch('helga.plugins.webhooks.WebhookRoot')
    @patch('helga.plugins.webhooks.reactor')
    def test_start_with_existing_root(self, reactor, WebhookRoot):
        self.plugin.root = Mock()
        self.plugin._start(Mock())
        assert not WebhookRoot.called

    def test_stop(self):
        tcp_mock = Mock()
        self.plugin.tcp = tcp_mock
        self.plugin._stop()

        assert tcp_mock.stopListening.called
        assert tcp_mock.loseConnection.called
        assert self.plugin.tcp is None

    def test_list_routes(self):
        client = Mock()
        root = Mock()
        root.routes = {
            '/foo/bar/': [['POST', 'GET'], lambda: None],
            u'/unicode/support/☃': [['PUT'], lambda: None]
        }

        self.plugin.root = root
        self.plugin.list_routes(client, 'me')

        call_list = client.msg.call_args_list
        assert call('me', 'me, here are the routes I know about') in call_list
        assert call('me', '[POST,GET] /foo/bar/') in call_list
        assert call('me', u'[PUT] /unicode/support/☃') in call_list

    def test_control_stop(self):
        # When running
        self.plugin.tcp = 'foo'
        with patch.object(self.plugin, '_stop') as stop:
            assert self.plugin.control('stop') == 'Webhooks service stopped'
            assert stop.called

        # When not running
        self.plugin.tcp = None
        with patch.object(self.plugin, '_stop') as stop:
            assert self.plugin.control('stop') == 'Webhooks service not running'
            assert not stop.called

    def test_control_start(self):
        # When running
        self.plugin.tcp = 'foo'
        with patch.object(self.plugin, '_start') as start:
            assert self.plugin.control('start') == 'Webhooks service already running'
            assert not start.called

        # When not running
        self.plugin.tcp = None
        with patch.object(self.plugin, '_start') as start:
            assert self.plugin.control('start') == 'Webhooks service started'
            assert start.called

    def test_run_defaults_to_list_routes(self):
        client = Mock()
        with patch.object(self.plugin, 'list_routes') as routes:
            self.plugin.run(client, '#bots', 'me', 'msg', 'cmd', [])
            client.me.assert_called_with('#bots', 'whispers to me')
            routes.assert_called_with(client, 'me')

    def test_run_list_routes(self):
        client = Mock()
        with patch.object(self.plugin, 'list_routes') as routes:
            self.plugin.run(client, '#bots', 'me', 'msg', 'cmd', ['routes'])
            client.me.assert_called_with('#bots', 'whispers to me')
            routes.assert_called_with(client, 'me')

    def test_run_start_stop_requires_operator(self):
        client = Mock(operators=[])
        with patch.object(self.plugin, 'control') as control:
            resp = self.plugin.run(client, '#bots', 'me', 'msg', 'cmd', ['start'])
            assert resp == 'Sorry me, Only an operator can do that'
            assert not control.called

            resp = self.plugin.run(client, '#bots', 'me', 'msg', 'cmd', ['stop'])
            assert resp == 'Sorry me, Only an operator can do that'
            assert not control.called

    def test_run_start_stop_as_operator(self):
        client = Mock(operators=['me'])
        with patch.object(self.plugin, 'control') as control:
            self.plugin.run(client, '#bots', 'me', 'msg', 'cmd', ['start'])
            control.assert_called_with('start')

            control.reset_mock()
            self.plugin.run(client, '#bots', 'me', 'msg', 'cmd', ['stop'])
            control.assert_called_with('stop')


class WebhookRootTestCase(TestCase):

    def setUp(self):
        self.client = Mock()
        self.root = webhooks.WebhookRoot(self.client)

    def test_render_returns_404(self):
        mock_fn = Mock(return_value='foo')
        request = Mock(path='/foo/bar/baz', method='POST')
        self.root.routes['/path/to/resource'] = (['GET'], mock_fn)

        assert '404 Not Found' == self.root.render(request)
        request.setResponseCode.assert_called_with(404)

    def test_render_returns_405(self):
        mock_fn = Mock(return_value='foo')
        request = Mock(path='/path/to/resource', method='POST')
        self.root.routes['/path/to/resource'] = (['GET'], mock_fn)

        assert '405 Method Not Allowed' == self.root.render(request)
        request.setResponseCode.assert_called_with(405)

    def test_render(self):
        mock_fn = Mock(return_value='foo')
        request = Mock(path='/path/to/resource', method='GET')
        self.root.routes['/path/to/resource'] = (['GET'], mock_fn)

        assert 'foo' == self.root.render(request)
        mock_fn.assert_called_with(request, self.root.irc_client)
        request.setHeader.assert_called_with('Server', 'helga')

    def test_reunder_handles_http_error(self):
        mock_fn = Mock(side_effect=webhooks.HttpError(404, 'foo not found'))
        request = Mock(path='/path/to/resource', method='GET')
        self.root.routes['/path/to/resource'] = (['GET'], mock_fn)

        assert 'foo not found' == self.root.render(request)
        request.setResponseCode.assert_called_with(404)

    def test_add_route(self):
        fn = lambda: None
        methods = ['GET', 'POST']
        path = '/path/to/resource'
        self.root.add_route(fn, path, methods)
        assert self.root.routes[path] == (methods, fn)



import time

from functools import partial

from helga.plugins.dubstep import dubstep


def test_dubstep_gives_wubs():
    assert 'wubwub' in dubstep('', '', '', '', '')


def test_dubstep_stops_after_max():
    dubstep._last = time.time()
    dubstep._counts['#bots'] = 0

    run = partial(dubstep, '', '#bots', '', '', '')

    assert 'wubwub' in run()
    assert 'wubwub' in run()
    assert 'wubwub' in run()
    assert 'STOP' in run()

    # Should start over now
    assert 'wubwub' in run()


def test_dubstep_after_timeout_resets_count():
    dubstep._last = time.time() - 30
    dubstep._counts['#bots'] = 10
    dubstep('', '#bots', '', '', '')
    assert dubstep._counts['#bots'] == 1



from mock import patch

from helga.plugins import oneliner


@patch('helga.plugins.oneliner.RESPONSES', {'foo': 'bar'})
def test_find_response_finds_none():
    assert oneliner.find_response('do not match') is None


@patch('helga.plugins.oneliner.RESPONSES', {'foo': ('bar', 'baz')})
def test_find_response_returns_random_item():
    assert oneliner.find_response('foo') in ('bar', 'baz')


@patch('helga.plugins.oneliner.RESPONSES', {'foo': 'bar'})
def test_find_response():
    assert oneliner.find_response('foo') == 'bar'



#CHANNEL_LOGGING = True
#CHANNEL_LOGGING_DIR = '/tmp/helga_channel_logs'
#CHANNEL_LOGGING_HIDE_CHANNELS = ['#foo']
#WEBHOOKS_CREDENTIALS = [('sduncan', 'sduncan')]
#ENABLED_WEBHOOKS = ['logger']
#COMMAND_ARGS_SHLEX = True

#SERVER = {
#    'HOST': 'chat.hipchat.com',
#    'PORT': 5222,
#    'JID': '291857_1819242@chat.hipchat.com',
#    'PASSWORD': 'One19@3#2',
#    'MUC_HOST': 'conf.hipchat.com',
#    'TYPE': 'xmpp',
#    'MANAGE_ROSTER': True,
#}
#CHANNELS = [
#    ('#291857_helgabot-dev',),
#]
#NICK = 'Helga Bot'
#COMMAND_PREFIX_BOTNICK = '@?HelgaBot'
#HIPCHAT_API_TOKEN = 'sIdbGFT7X608xZDhn0VNAdKrqwuOefg7DXdR1cdJ'

#SERVER = {
#    'HOST': 'helga.local',
#    'PORT': 5222,
#    'USERNAME': 'helga',
#    'PASSWORD': 'helga',
#    'TYPE': 'xmpp',
#}

SERVER = {
    'HOST': 'hipchat.pdrop.net',
    'PORT': 5222,
    'JID': '1_11@chat.btf.hipchat.com',
    'PASSWORD': 'One19@3#2',
    'MUC_HOST': 'conf.btf.hipchat.com',
    'TYPE': 'xmpp',
}
CHANNELS = [
    ('#1_xcmg',),
]
NICK = 'Shaun Duncan'
COMMAND_PREFIX_BOTNICK = '@?ShaunDuncan'
HIPCHAT_API_TOKEN = 'sK7EBKb7exckFIXBZ89BJ13lCrRXPtTAaagPG0PD'
HIPCHAT_API_ENDPOINT = 'https://hipchat.pdrop.net/api'



import datetime
import re

import freezegun
import pytest

from mock import patch, Mock

from helga import log


@patch('helga.log.logging')
@patch('helga.log.settings')
def test_getLogger(settings, logging):
    logger = Mock()
    handler = Mock()
    formatter = Mock()

    settings.LOG_LEVEL = 'INFO'
    settings.LOG_FILE = None
    settings.LOG_FORMAT = None

    expected_format = '%(asctime)-15s [%(levelname)s] [%(name)s:%(lineno)d]: %(message)s'

    logging.getLogger.return_value = logger
    logging.StreamHandler.return_value = handler
    logging.Formatter.return_value = formatter

    log.getLogger('foo')
    logging.getLogger.assert_called_with('foo')
    logger.setLevel.assert_called_with(logging.INFO)
    logger.addHandler.assert_called_with(handler)
    logging.Formatter.assert_called_with(expected_format)
    handler.setFormatter.assert_called_with(formatter)
    assert not logger.propagate


@patch('helga.log.logging')
@patch('helga.log.settings')
def test_get_logger_uses_log_file(settings, logging):
    logger = Mock()
    handler = Mock()
    formatter = Mock()

    settings.LOG_LEVEL = 'INFO'
    settings.LOG_FILE = '/path/to/foo.log'
    settings.LOG_FORMAT = None

    logging.getLogger.return_value = logger
    logging.StreamHandler.return_value = handler
    logging.Formatter.return_value = formatter

    log.getLogger('foo')
    logging.handlers.RotatingFileHandler.assert_called_with(filename=settings.LOG_FILE,
                                                            maxBytes=50*1024*1024,
                                                            backupCount=6)


@patch('helga.log.logging')
@patch('helga.log.settings')
def test_get_logger_uses_custom_formatter(settings, logging):
    logger = Mock()
    handler = Mock()
    formatter = Mock()

    settings.LOG_LEVEL = 'INFO'
    settings.LOG_FILE = '/path/to/foo.log'
    settings.LOG_FORMAT = 'blah blah blah'

    logging.getLogger.return_value = logger
    logging.StreamHandler.return_value = handler
    logging.Formatter.return_value = formatter

    log.getLogger('foo')
    logging.Formatter.assert_called_with(settings.LOG_FORMAT)


@patch('helga.log.os')
@patch('helga.log.logging')
@patch('helga.log.settings')
def test_get_channel_logger(settings, logging, os):
    logger = Mock()
    handler = Mock()
    formatter = Mock()

    def os_join(*args):
        return '/'.join(args)

    settings.CHANNEL_LOGGING_DIR = '/path/to/channels'
    settings.CHANNEL_LOGGING_DB = False
    os.path.exists.return_value = True
    os.path.join = os_join

    # Mocked returns
    logging.getLogger.return_value = logger
    logging.Formatter.return_value = formatter

    with patch.object(log, 'ChannelLogFileHandler'):
        log.ChannelLogFileHandler.return_value = handler
        log.get_channel_logger('#foo')

        # Gets the right logger
        logging.getLogger.assert_called_with('channel_logger/#foo')
        logger.setLevel.assert_called_with(logging.INFO)
        assert logger.propagate is False

        # Sets the handler correctly
        log.ChannelLogFileHandler.assert_called_with('/path/to/channels/#foo')
        handler.setFormatter.assert_called_with(formatter)

        # Sets the formatter correctly
        logging.Formatter.assert_called_with('%(utctime)s - %(nick)s - %(message)s')

        # Logger uses the handler
        logger.addHandler.assert_called_with(handler)


@patch('helga.log.os')
@patch('helga.log.logging')
@patch('helga.log.settings')
def test_get_channel_logger_creates_log_dirs(settings, logging, os):
    def os_join(*args):
        return '/'.join(args)

    settings.CHANNEL_LOGGING_DIR = '/path/to/channels'
    settings.CHANNEL_LOGGING_DB = False
    os.path.exists.return_value = False
    os.path.join = os_join

    log.get_channel_logger('#foo')

    os.makedirs.assert_called_with('/path/to/channels/#foo')


class TestChannelLogFileHandler(object):

    def setup(self):
        self.handler = log.ChannelLogFileHandler('/tmp')

    def test_setup_correctly(self):
        assert self.handler.basedir == '/tmp'
        assert re.match(r'/tmp/[0-9]{4}-[0-9]{2}-[0-9]{2}.txt',
                        self.handler.baseFilename)

    def test_compute_next_rollover(self):
        expected = datetime.datetime(2014, 11, 1)
        with freezegun.freeze_time('2014-10-31 08:15'):
            assert self.handler.compute_next_rollover() == expected

    @freezegun.freeze_time('2014-10-31 08:15')
    def test_current_filename(self):
        assert self.handler.current_filename() == '2014-10-31.txt'

    @pytest.mark.parametrize('datestr,rollover', [
        ('2014-10-31 08:15', False),
        ('2014-11-01 00:00', True),
        ('2014-11-01 08:15', True),
    ])
    def test_shouldRollover(self, datestr, rollover):
        self.handler.next_rollover = datetime.datetime(2014, 11, 1)
        with freezegun.freeze_time(datestr):
            assert self.handler.shouldRollover(None) is rollover

    def test_do_rollover(self):
        stream = Mock()
        self.handler.stream = stream

        old_rollover = self.handler.next_rollover
        old_filename = self.handler.baseFilename
        expected_rollover = datetime.datetime(2014, 11, 1, 0, 0, 0)

        assert old_rollover != expected_rollover

        with freezegun.freeze_time('2014-10-31 08:15'):
            with patch.object(self.handler, '_open'):
                self.handler.doRollover()
                assert stream.close.called
                assert self.handler._open.called
                assert self.handler.baseFilename != old_filename
                assert self.handler.baseFilename == '/tmp/2014-10-31.txt'
                assert self.handler.next_rollover == expected_rollover


def test_utc_time_filter():
    record = Mock()
    filter = log.UTCTimeLogFilter()
    date = datetime.datetime(2014, 10, 31, 8, 15)
    with freezegun.freeze_time(date):
        filter.filter(record)
        assert record.utcnow == date
        assert record.utctime == '08:15:00'



from __future__ import absolute_import

import argparse
import os

import smokesignal

from twisted.internet import reactor, ssl

from helga import settings


def _get_comm():  # pragma: no cover
    # XXX: This is here ONLY to make patching eaiser. We have to delay the import
    # of comm so that settings are properly overridden for any class attributes
    from helga import comm
    return comm


def run():
    """
    Run the helga process
    """
    comm = _get_comm()
    smokesignal.emit('started')

    factory = comm.Factory()
    if settings.SERVER.get('SSL', False):
        reactor.connectSSL(settings.SERVER['HOST'],
                           settings.SERVER['PORT'],
                           factory,
                           ssl.ClientContextFactory())
    else:
        reactor.connectTCP(settings.SERVER['HOST'],
                           settings.SERVER['PORT'],
                           factory)
    reactor.run()


def main():
    """
    Main entry point for the helga console script
    """
    parser = argparse.ArgumentParser(description='The helga IRC bot')
    parser.add_argument('--settings', help=(
        'Custom helga settings overrides. This should be an importable python module '
        'like "foo.bar.baz" or a path to a settings file like "path/to/settings.py". '
        'This can also be set via the HELGA_SETTINGS environment variable, however '
        'this flag takes precedence.'
    ))
    args = parser.parse_args()

    settings_file = os.environ.get('HELGA_SETTINGS', '')

    if args.settings:
        settings_file = args.settings

    settings.configure(settings_file)
    run()



"""
Twisted protocol and communication implementations for IRC
"""
import time

from collections import defaultdict

import smokesignal

from twisted.internet import protocol, reactor
from twisted.words.protocols import irc

from helga import settings, log
from helga.plugins import registry
from helga.util import encodings


logger = log.getLogger(__name__)


class Factory(protocol.ClientFactory):
    """
    The client factory for twisted. Ensures that a client is properly created and handles
    auto reconnect if helga is configured for it (see settings :data:`~helga.settings.AUTO_RECONNECT`
    and :data:`~helga.settings.AUTO_RECONNECT_DELAY`)
    """

    def buildProtocol(self, address):
        """
        Build the helga protocol for twisted, or in other words, create the client
        object and return it.

        :param address: an implementation of :class:`twisted.internet.interfaces.IAddress`
        :returns: an instance of :class:`helga.comm.Client`
        """
        logger.debug('Constructing Helga protocol')
        return Client(factory=self)

    def clientConnectionLost(self, connector, reason):
        """
        Handler for when the IRC connection is lost. Handles auto reconnect if helga
        is configured for it (see settings :data:`~helga.settings.AUTO_RECONNECT` and
        :data:`~helga.settings.AUTO_RECONNECT_DELAY`)
        """
        logger.info('Connection to server lost: %s', reason)

        # FIXME: Max retries
        if getattr(settings, 'AUTO_RECONNECT', True):
            delay = getattr(settings, 'AUTO_RECONNECT_DELAY', 5)
            reactor.callLater(delay, connector.connect)
        else:
            raise reason

    def clientConnectionFailed(self, connector, reason):
        """
        Handler for when the IRC connection fails. Handles auto reconnect if helga
        is configured for it (see settings :data:`~helga.settings.AUTO_RECONNECT` and
        :data:`~helga.settings.AUTO_RECONNECT_DELAY`)
        """
        logger.warning('Connection to server failed: %s', reason)

        # FIXME: Max retries
        if getattr(settings, 'AUTO_RECONNECT', True):
            delay = getattr(settings, 'AUTO_RECONNECT_DELAY', 5)
            reactor.callLater(delay, connector.connect)
        else:
            reactor.stop()


class Client(irc.IRCClient):
    """
    An implementation of :class:`twisted.words.protocols.irc.IRCClient` with some overrides
    derived from helga settings (see :ref:`config`). Some methods are overridden
    to provide additional functionality.

    .. attribute:: channels
        :annotation: = set()

        A set containing all of the channels the bot is currently in

    .. attribute:: operators
        :annotation: = set()

        A set containing all of the configured operators (setting :data:`~helga.settings.OPERATORS`)

    .. attribute:: last_message
        :annotation: = dict()

        A channel keyed dictionary containing dictionaries of nick -> message of the last messages
        the bot has seen a user send on a given channel. For instance, if in the channel ``#foo``::

            <sduncan> test

        The contents of this dictionary would be::

            self.last_message['#foo']['sduncan'] = 'test'

    .. attribute:: channel_loggers
        :annotation: = dict()

        A dictionary of known channel loggers, keyed off the channel name
    """

    #: The preferred IRC nick of the bot instance (setting :data:`~helga.settings.NICK`)
    nickname = None

    #: A username should the IRC server require authentication (setting :data:`~helga.settings.SERVER`)
    username = None

    #: A password should the IRC server require authentication (setting :data:`~helga.settings.SERVER`)
    password = None

    #: An integer, in seconds, if IRC messages should be sent at a limit of once per this many seconds.
    #: ``None`` implies no limit. (setting :data:`~helga.settings.RATE_LIMIT`)
    lineRate = None

    #: The URL where the source of the bot is found
    sourceURL = 'http://github.com/shaunduncan/helga'

    #: The assumed encoding of IRC messages
    encoding = 'UTF-8'

    #: A backup nick should the preferred :attr:`nickname` be taken. This defaults to a string in the
    #: form of the preferred nick plus the timestamp when the process was started (i.e. helga_12345)
    erroneousNickFallback = None

    def __init__(self, factory=None):
        self.factory = factory
        self.erroneousNickFallback = '{0}_{1}'.format(settings.NICK, int(time.time()))

        # These are set here to ensure using properly overridden settings
        self.nickname = settings.NICK
        self.username = settings.SERVER.get('USERNAME', None)
        self.password = settings.SERVER.get('PASSWORD', None)
        self.lineRate = getattr(settings, 'RATE_LIMIT', None)

        # Pre-configured helga admins
        self.operators = set(getattr(settings, 'OPERATORS', []))

        # Things to keep track of
        self.channels = set()
        self.last_message = defaultdict(dict)  # Dict of x[channel][nick]
        self.channel_loggers = {}

    def get_channel_logger(self, channel):
        """
        Gets a channel logger, keeping track of previously requested ones.
        (see :ref:`builtin.channel_logging`)

        :param str channel: A channel name
        """
        if channel not in self.channel_loggers:
            self.channel_loggers[channel] = log.get_channel_logger(channel)
        return self.channel_loggers[channel]

    def log_channel_message(self, channel, nick, message):
        """
        Logs one or more messages by a user on a channel using a channel logger.
        If channel logging is not enabled, nothing happens. (see :ref:`builtin.channel_logging`)

        :param str channel: A channel name
        :param str nick: The nick of the user sending an IRC message
        :param str message: The IRC message
        """
        if not settings.CHANNEL_LOGGING:
            return
        chan_logger = self.get_channel_logger(channel)
        chan_logger.info(message, extra={'nick': nick})

    def connectionMade(self):
        logger.info('Connection made to %s', settings.SERVER['HOST'])
        irc.IRCClient.connectionMade(self)

    @encodings.from_unicode_args
    def connectionLost(self, reason):
        logger.info('Connection to %s lost: %s', settings.SERVER['HOST'], reason)
        irc.IRCClient.connectionLost(self, reason)

    def signedOn(self):
        """
        Called when the client has successfully signed on to IRC. Establishes automatically
        joining channels. Sends the ``signon`` signal (see :ref:`plugins.signals`)
        """
        for channel in settings.CHANNELS:
            # If channel is more than one item tuple, second value is password
            if isinstance(channel, (tuple, list)):
                self.join(*channel)
            else:
                self.join(channel)
        smokesignal.emit('signon', self)

    def joined(self, channel):
        """
        Called when the client successfully joins a new channel. Adds the channel to the known
        channel list and sends the ``join`` signal (see :ref:`plugins.signals`)

        :param str channel: the channel that has been joined
        """
        logger.info('Joined %s', channel)
        self.channels.add(channel)
        smokesignal.emit('join', self, channel)

    def left(self, channel):
        """
        Called when the client successfully leaves a channel. Removes the channel from the known
        channel list and sends the ``left`` signal (see :ref:`plugins.signals`)

        :param str channel: the channel that has been left
        """
        logger.info('Left %s', channel)
        self.channels.discard(channel)
        smokesignal.emit('left', self, channel)

    def parse_nick(self, full_nick):
        """
        Parses a nick from a full IRC user string. For example from ``me!~myuser@localhost``
        would return ``me``.

        :param str full_nick: the full IRC user string of the form ``{nick}!~{user}@{host}``
        :returns: The nick portion of the IRC user string
        """
        return full_nick.split('!')[0]

    def is_public_channel(self, channel):
        """
        Checks if a given channel is public or not. A channel is public if it starts with
        '#' and is not the bot's nickname (which occurs when a private message is received)

        :param str channel: the channel name to check
        """
        return self.nickname != channel and channel.startswith('#')

    @encodings.to_unicode_args
    def privmsg(self, user, channel, message):
        """
        Handler for an IRC message. This method handles logging channel messages (if it occurs
        on a public channel) as well as allowing the plugin manager to send the message to all
        registered plugins. Should the plugin manager yield a response, it will be sent back
        over IRC.

        :param str user: IRC user string of the form ``{nick}!~{user}@{host}``
        :param str channel: the channel from which the message came
        :param str message: the message contents
        """
        user = self.parse_nick(user)
        message = message.strip()

        # Log the incoming message and notify message subscribers
        logger.debug('[<--] %s/%s - %s', channel, user, message)
        is_public = self.is_public_channel(channel)

        # When we get a priv msg, the channel is our current nick, so we need to
        # respond to the user that is talking to us
        if is_public:
            # Only log convos on public channels
            self.log_channel_message(channel, user, message)
        else:
            channel = user

        # Some things should go first
        try:
            channel, user, message = registry.preprocess(self, channel, user, message)
        except (TypeError, ValueError):
            pass

        # if not message.has_response:
        responses = registry.process(self, channel, user, message)

        if responses:
            message = u'\n'.join(responses)
            self.msg(channel, message)

            if is_public:
                self.log_channel_message(channel, self.nickname, message)

        # Update last message
        self.last_message[channel][user] = message

    def alterCollidedNick(self, nickname):
        """
        Called when the both has a nickname collision. This will generate a new nick
        containing the perferred nick and the current timestamp.

        :param str nickname: the nickname that was already taken
        """
        logger.info('Nick %s already taken', nickname)

        parts = nickname.split('_')
        if len(parts) > 1:
            parts = parts[:-1]

        stripped = '_'.join(parts)
        self.nickname = '{0}_{1}'.format(stripped, int(time.time()))

        return self.nickname

    def kickedFrom(self, channel, kicker, message):
        logger.warning('%s kicked bot from %s: %s', kicker, channel, message)
        self.channels.discard(channel)

    @encodings.from_unicode_args
    def msg(self, channel, message):
        """
        Send a message over IRC to the specified channel

        :param channel: The IRC channel to send the message to. A channel not prefixed by a '#'
                        will be sent as a private message to a user with that nick.
        :param message: The message to send
        """
        logger.debug('[-->] %s - %s', channel, message)
        irc.IRCClient.msg(self, channel, message)

    def on_invite(self, inviter, invitee, channel):
        """
        Handler for /INVITE commands. If the invitee is the bot, it will join the requested channel.

        :param str inviter: IRC user string of the form ``{nick}!~{user}@{host}``
        :param str invitee: the nick of the user receiving the invite
        :param str channel: the channel to which invitee has been invited
        """
        nick = self.parse_nick(inviter)
        if invitee == self.nickname:
            logger.info('%s invited %s to %s', nick, invitee, channel)
            self.join(channel)

    def irc_unknown(self, prefix, command, params):
        """
        Handler for any unknown IRC commands. Currently handles /INVITE commands

        :param str prefix: any command prefix, such as the IRC user
        :param str command: the IRC command received
        :param list params: list of parameters for the given command
        """
        if command.lower() == 'invite':
            self.on_invite(prefix, params[0], params[1])

    @encodings.from_unicode_args
    def me(self, channel, message):
        """
        Equivalent to: /me message

        :param channel: The IRC channel to send the message to. A channel not prefixed by a '#'
                        will be sent as a private message to a user with that nick.
        :param message: The message to send
        """
        # A proxy for the WTF-named method `describe`. Basically the same as doing `/me waves`
        irc.IRCClient.describe(self, channel, message)

    def userJoined(self, user, channel):
        """
        Called when a user joins a channel in which the bot resides. Responsible for sending
        the ``user_joined`` signal (see :ref:`plugins.signals`)

        :param str user: IRC user string of the form ``{nick}!~{user}@{host}``
        :param str channel: the channel in which the event occurred
        """
        nick = self.parse_nick(user)
        smokesignal.emit('user_joined', self, nick, channel)

    def userLeft(self, user, channel):
        """
        Called when a user leaves a channel in which the bot resides. Responsible for sending
        the ``user_left`` signal (see :ref:`plugins.signals`)

        :param str user: IRC user string of the form ``{nick}!~{user}@{host}``
        :param str channel: the channel in which the event occurred
        """
        nick = self.parse_nick(user)
        smokesignal.emit('user_left', self, nick, channel)

    @encodings.from_unicode_args
    def join(self, channel, key=None):
        """
        Join a channel, optionally with a passphrase required to join.

        :param str channel: the name of the channel to join
        :param str key: an optional passphrase used to join the given channel
        """
        logger.info("Joining channel %s", channel)
        irc.IRCClient.join(self, channel, key=key)

    @encodings.from_unicode_args
    def leave(self, channel, reason=None):
        """
        Leave a channel, optionally with a reason for leaving

        :param str channel: the name of the channel to leave
        :param str reason: an optional reason for leaving
        """
        logger.info("Leaving channel %s: %s", channel, reason)
        irc.IRCClient.leave(self, channel, reason=reason)



import random
import re

import requests
import smokesignal

from requests.auth import HTTPBasicAuth
from twisted.internet import reactor

from helga import log, settings
from helga.db import db
from helga.plugins import command, match, ACKS, ResponseNotReady
from helga.util.encodings import to_unicode


logger = log.getLogger(__name__)


# These are initialized on client signon
JIRA_PATTERNS = set()


@smokesignal.on('signon')
def init_jira_patterns(*args, **kwargs):
    """
    Signal callback for IRC signon. This pulls down and caches all the stored
    JIRA ticket patterns so we don't have to do it on every message received
    """
    global JIRA_PATTERNS

    if db is None:  # pragma: no cover
        logger.warning('Cannot initialize JIRA patterns. No database connection')

    JIRA_PATTERNS = set(item['re'] for item in db.jira.find())


def find_jira_numbers(message):
    """
    Finds all jira ticket numbers in a message. This will ignore any that already
    appear in a URL
    """
    global JIRA_PATTERNS

    pat = r'(https?://.*?)?(({0})-\d+)($|[\s\W]+)'.format('|'.join(JIRA_PATTERNS))
    tickets = []

    if not JIRA_PATTERNS:
        return tickets

    for matched in re.findall(pat, message, re.IGNORECASE):
        # if matched[0] is not empty, the ticket is in a URL. Ignore it
        if matched[0]:
            continue
        tickets.append(matched[1])

    return tickets


def add_re(pattern):
    """
    Adds a ticket pattern from the database and local cache
    """
    global JIRA_PATTERNS

    if pattern not in JIRA_PATTERNS:
        logger.info('Adding new JIRA ticket RE: %s', pattern)
        JIRA_PATTERNS.add(pattern)
        re_doc = {'re': pattern}

        # Store in DB
        if not db.jira.find(re_doc).count():
            db.jira.insert(re_doc)
    else:  # pragma: no cover
        logger.info('JIRA ticket RE already exists: %s', pattern)

    return random.choice(ACKS)


def remove_re(pattern):
    """
    Removes a ticket pattern from the database and local cache
    """
    global JIRA_PATTERNS

    logger.info('Removing JIRA ticket RE: %s', pattern)
    JIRA_PATTERNS.discard(pattern)
    db.jira.remove({'re': pattern})

    return random.choice(ACKS)


def jira_command(client, channel, nick, message, cmd, args):
    """
    Command handler for the jira plugin
    """
    try:
        subcmd, pattern = args[:2]
    except ValueError:
        return None

    if subcmd == 'add_re':
        return add_re(pattern)

    if subcmd == 'remove_re':
        return remove_re(pattern)

    return None


def _rest_desc(ticket, url, auth=None):
    api_url = to_unicode(settings.JIRA_REST_API)
    resp = requests.get(api_url.format(ticket=ticket), auth=auth)

    try:
        resp.raise_for_status()
    except:
        logger.error('Error getting JIRA ticket %s. Status %s', ticket, resp.status_code)
        return

    try:
        return u'[{0}] {1} ({2})'.format(ticket.upper(), resp.json()['fields']['summary'], url)
    except:
        return u'[{0}] {1}'.format(ticket.upper(), url)


def jira_full_descriptions(client, channel, urls):
    """
    Meant to be run asynchronously because it uses the network
    """
    descriptions = []
    user_pass = getattr(settings, 'JIRA_AUTH', ('', ''))

    if all(user_pass):
        auth = HTTPBasicAuth(*user_pass)
    else:
        auth = None

    for ticket, url in urls.iteritems():
        desc = _rest_desc(ticket, url, auth)
        if desc is not None:
            descriptions.append(desc)

    if descriptions:
        client.msg(channel, '\n'.join(descriptions))


def jira_match(client, channel, nick, message, matches):
    jira_url = to_unicode(settings.JIRA_URL)
    full_urls = dict((s, jira_url.format(ticket=s)) for s in matches)

    if not getattr(settings, 'JIRA_SHOW_FULL_DESCRIPTION', True):
        return u'{0} might be talking about JIRA ticket: {1}'.format(nick, ', '.join(full_urls.values()))

    # Otherwise, do the fetching with a deferred
    reactor.callLater(0, jira_full_descriptions, client, channel, full_urls)
    raise ResponseNotReady


@match(find_jira_numbers)
@command('jira', help="Add or remove jira ticket patterns, excludeing numbers."
                      "Usage: helga jira (add_re|remove_re) <pattern>")
def jira(client, channel, nick, message, *args):
    """
    A plugin for showing URLs to JIRA ticket numbers. This is both a command to add or remove
    patterns, and a match to automatically show them. The match requires a setting JIRA_URL
    which must contain a ``{ticket}`` substring. For example, ``http://localhost/{ticket}``.

    The command takes a pattern as an argument, minus any numbers. For example, if there are JIRA
    tickets like FOOBAR-1, FOOBAR-2, and FOOBAR-3. Then you could manage the pattern via::

        helga jira add_re FOOBAR
        helga jira remove_re FOOBAR

    Ticket numbers are automatically detected.
    """
    if len(args) == 2:
        fn = jira_command
    else:
        fn = jira_match
    return fn(client, channel, nick, message, *args)



import random
import time

from collections import defaultdict

from helga.plugins import match


MAX_WUBS = 3
WUB_TIMEOUT = 10


@match('dubstep', priority=0)
def dubstep(client, channel, nick, message, matches):
    """
    Dubstep can be described as a rapid succession of wub wubs, wow wows, and yep yep yep yeps
    """
    now = time.time()

    if dubstep._last and (now - dubstep._last) > WUB_TIMEOUT:
        dubstep._counts[channel] = 0

    dubstep._last = now
    if dubstep._counts[channel] >= MAX_WUBS:
        dubstep._counts[channel] = 0
        return u'STOP! MY HEAD IS VIBRATING'
    else:
        dubstep._counts[channel] += 1
        return u'wubwub' * dubstep._counts[channel] * random.randint(1, 4)

dubstep._counts = defaultdict(int)
dubstep._last = None



# -*- coding: utf8 -*-
import types

from mock import Mock
from unittest import TestCase

from helga import settings
from helga.webhooks.announcements import announce


class AnnouncementTestCase(TestCase):

    def setUp(self):
        self.client = Mock()
        self.request = Mock(args={})

        def _set_response_code(self, code):
            self.response_code = code

        self.request.setResponseCode = types.MethodType(_set_response_code, self.request)

        # Ensure requests are always authenticated
        self.request.getUser.return_value = 'user'
        self.request.getPassword.return_value = 'password'
        settings.WEBHOOKS_CREDENTIALS = [('user', 'password')]

    def test_requires_message_content(self):
        assert 'Param message is required' == announce(self.request, self.client, '#foo')
        assert self.request.response_code == 400

    def test_formatted_channel(self):
        self.request.args['message'] = ['bar']
        assert 'Message Sent' == announce(self.request, self.client, 'foo')
        self.client.msg.assert_called_with('#foo', 'bar')

    def test_handles_unicode(self):
        snowman = u'☃'
        self.request.args['message'] = [snowman]
        assert 'Message Sent' == announce(self.request, self.client, '#foo')
        self.client.msg.assert_called_with('#foo', snowman)

    def test_announce(self):
        self.request.args['message'] = ['bar']
        assert 'Message Sent' == announce(self.request, self.client, '#foo')
        self.client.msg.assert_called_with('#foo', 'bar')



import functools
import pkg_resources
import subprocess
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

from pip.req import parse_requirements as parse_reqs

import helga


# Compatibility with older versions of pip
pip_dist = pkg_resources.get_distribution('pip')
pip_version = tuple(map(int, pip_dist.version.split('.')))

# Use a base partial that will be updated depending on the version of pip
parse_requirements = functools.partial(parse_reqs, options=None)

if pip_version < (1, 2):
    # pip versions before 1.2 require an options keyword for using it outside
    # of invoking a pip shell command
    from pip.baseparser import parser
    parse_requirements.keywords['options'] = parser.parse_args()[0]

if pip_version >= (1, 5):
    # pip 1.5 introduced a session kwarg that is required in later versions
    from pip.download import PipSession
    parse_requirements.keywords['session'] = PipSession()


# If installing on python 2.6, we need to install the argparse backport
extra_requires = []
if sys.version_info[:2] == (2, 6):
    extra_requires = ['argparse==1.3.0']


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        return subprocess.call('tox')


setup(name=helga.__title__,
      version=helga.__version__,
      description=helga.__description__,
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Topic :: Communications :: Chat :: Internet Relay Chat',
          'Framework :: Twisted',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      keywords='irc bot helga',
      author=helga.__author__,
      author_email='shaun.duncan@gmail.com',
      url='https://github.com/shaunduncan/helga',
      license=helga.__license__,
      packages=find_packages(),
      package_data={
          'helga': ['webhooks/logger/*.mustache'],
      },
      install_requires=[
          str(req.req) for req in parse_requirements('requirements.txt')
      ] + extra_requires,
      tests_require=[
          'freezegun',
          'mock',
          'pretend',
          'tox',
          'pytest',
      ],
      cmdclass={'test': PyTest},
      entry_points=dict(
          helga_plugins=[
              'dubstep      = helga.plugins.dubstep:dubstep',
              'facts        = helga.plugins.facts:facts',
              'giphy        = helga.plugins.giphy:giphy',
              'help         = helga.plugins.help:help',
              'icanhazascii = helga.plugins.icanhazascii:icanhazascii',
              'jira         = helga.plugins.jira:jira',
              'loljava      = helga.plugins.loljava:make_bullshit_java_thing',
              'manager      = helga.plugins.manager:manager',
              'meant_to_say = helga.plugins.meant_to_say:meant_to_say',
              'no_more_olga = helga.plugins.no_more_olga:no_more_olga',
              'oneliner     = helga.plugins.oneliner:oneliner',
              'operator     = helga.plugins.operator:operator',
              'poems        = helga.plugins.poems:poems',
              'reminders    = helga.plugins.reminders:reminders',
              'reviewboard  = helga.plugins.reviewboard:reviewboard',
              'stfu         = helga.plugins.stfu:stfu',
              'webhooks     = helga.plugins.webhooks:WebhookPlugin',
              'wiki_whois   = helga.plugins.wiki_whois:wiki_whois',
          ],
          helga_webhooks=[
              'announcements = helga.webhooks.announcements:announce',
              'logger        = helga.webhooks.logger:logger'
          ],
          console_scripts=[
              'helga = helga.bin.helga:main',
          ],
      ),
)



from helga import log
from helga.plugins.webhooks import authenticated, route


logger = log.getLogger(__name__)


@route('/announce/(?P<channel>[\w\-_]+)', methods=['POST'])
@authenticated
def announce(request, irc_client, channel):
    """
    An endpoint for announcing a message on a channel. POST only, must
    provide a single data param 'message'
    """
    if not channel.startswith('#'):
        channel = '#{0}'.format(channel)

    message = request.args.get('message', [''])[0]
    if not message:
        request.setResponseCode(400)
        return 'Param message is required'

    logger.info('Sending message to %s: "%s"', channel, message)
    irc_client.msg(channel, message)

    # Return accepted
    return 'Message Sent'



from __future__ import unicode_literals

import random
import re
import time

from datetime import datetime

import pytz

from helga import log, settings
from helga.db import db
from helga.plugins import command, match, ACKS


logger = log.getLogger(__name__)


def term_regex(term):
    """
    Returns a case-insensitive regex for searching terms
    """
    return re.compile(r'^{0}$'.format(re.escape(term)), re.IGNORECASE)


def show_fact(term):
    """
    Shows a fact stored for a given term, using a case-insensitive search.
    If a fact has an author, it will be shown. If it has a timestamp, that will
    also be shown.
    """
    logger.info('Showing fact %s', term)
    record = db.facts.find_one({'term': term_regex(term)})

    if record is None:
        return None

    # If it isn't authored
    if not record.get('set_by', ''):
        return record['fact']

    if 'set_date' not in record:
        return '{fact} ({set_by})'.format(**record)

    # Otherwise, do normal formatting
    tz = getattr(settings, 'TIMEZONE', 'US/Eastern')
    try:
        timestamp = datetime.fromtimestamp(record['set_date'], tz=pytz.timezone(tz))
    except TypeError:
        timestamp = record['set_date'].replace(tzinfo=pytz.timezone(tz))
    record['fmt_dt'] = datetime.strftime(timestamp, '%m/%d/%Y %I:%M%p')

    return '{fact} ({set_by} on {fmt_dt})'.format(**record)


def add_fact(term, fact, author=''):
    """
    Records a new fact with a given term. Optionally can set an author
    """
    logger.info('Adding new fact %s: %s', term, fact)

    if not db.facts.find({'term': term_regex(term)}).count():
        db.facts.insert({
            'term': term,
            'fact': fact,
            'set_by': author,
            'set_date': time.time()
        })
        db.facts.ensure_index('term')


def forget_fact(term):
    """
    Forgets a fact by removing it from the database
    """
    logger.info('Removing fact %s', term)
    db.facts.remove({'term': term_regex(term)})
    return random.choice(ACKS)


def replace_fact(term, fact, author=''):
    """
    Replaces an existing fact by removing it, then adding the new definition
    """
    forget_fact(term)
    add_fact(term, fact, author)
    return random.choice(ACKS)


def facts_command(client, channel, nick, message, cmd, args):
    if cmd == 'forget':
        return forget_fact(' '.join(args))
    if cmd == 'replace':
        if '<with>' not in args:
            return 'No definition supplied.'
        all_args = ' '.join(args)
        term, fact = all_args.split(' <with> ', 1)
        return replace_fact(term, fact, author=nick)


def facts_match(client, channel, nick, message, found):
    parts = found[0]

    # Like: foo?
    if isinstance(parts, basestring):
        return show_fact(parts)

    # Allow customizing facts to be non-automatic and work more like commands
    if getattr(settings, 'FACTS_REQUIRE_NICKNAME', False):
        try:
            nonick = re.findall(r'^{0}\W*\s(.*)$'.format(client.nickname), parts[0])[0]
        except IndexError:
            logger.debug('Facts require the current bot nick. Ignoring: %s', parts[0])
            return None
        else:
            parts = [nonick] + list(parts[1:])

    if parts[2].strip() == '<reply>':
        author = ''
        fact = parts[-1]
    else:
        # This nasty join is to ignore the empty part for <reply>
        author = nick
        fact = ' '.join(parts[:2] + parts[-1:])

    return add_fact(parts[0], fact, author)


@command('forget', help='Forget a stored fact. Usage: <botnick> forget foo')
@match(r'^(.*?) (is|are)( <reply>\s*)?(.+)$')  # Storing facts
@match(r'^(.*)\?$')  # Showing facts
def facts(client, channel, nick, message, *args):
    """
    A plugin for helga to automatically remember important things. Unless specified
    by the setting ``FACTS_REQUIRE_NICKNAME``, facts are automatically stored when
    a user says: ``something is something else``. Otherwise, facts must be explicitly
    added: ``helga something is something else``.

    Response format is, by default, the full message that was sent, including an author
    and timestamp. However, if a user specifies the string '<reply>' in their message,
    then only the words that follow '<reply>' will be returned in the response.

    For example::

        <sduncan> foo is bar
        <sduncan> foo?
        <helga> foo is bar (sduncan on 12/02/2013)

    Or::

        <sduncan> foo is <reply> bar
        <sduncan> foo?
        <helga> bar (sduncan on 12/02/2013)

    To remove a fact, you must ask specifically using the ``forget`` command::

        <sduncan> helga forget foo
        <helga> forgotten

    To replace a fact, you must use the ``replace`` command, andprovide the
    term as well as the new definition, separated by  '<watch>'::

        <sduncan> helga replace foo <with> new def
        <helga> replaced
    """
    if len(args) == 2:
        return facts_command(client, channel, nick, message, *args)

    # Anything else is a match
    return facts_match(client, channel, nick, message, args[0])
